# June 2020, Lewis Gaul

"""
Simulate played games.
"""

__all__ = ("MinefieldWidget",)

import logging
from typing import Dict, List, Mapping, Optional

from PyQt5.QtCore import QTimer
from PyQt5.QtGui import QMouseEvent, QPixmap
from PyQt5.QtWidgets import (
    QDialog,
    QFrame,
    QGraphicsScene,
    QGraphicsView,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from ...shared.types import CellContents, CellImageType, Coord
from ..utils import CellUpdate_T
from ._base import update_cell_images


logger = logging.getLogger(__name__)


# TODO: Inherit from MinefieldWidgetBase
class MinefieldWidget(QDialog):
    def __init__(
        self,
        parent: Optional[QWidget],
        x_size: int,
        y_size: int,
        cell_updates: List[CellUpdate_T],
    ):
        super().__init__(parent)
        self._x_size = x_size
        self._y_size = y_size
        self._remaining_cell_updates = cell_updates

        self._cell_images: Dict[CellContents, QPixmap] = {}
        update_cell_images(
            self._cell_images, self.btn_size, {CellImageType.BUTTONS: "standard"}
        )

        self._scene = QGraphicsScene()
        self.setModal(True)
        self.setWindowTitle("Highscore replay")
        self._setup_ui()

        for c in [(x, y) for x in range(self.x_size) for y in range(self.y_size)]:
            self._set_cell_image(c, CellContents.Unclicked)

        self._animation_started = False

    @property
    def x_size(self) -> int:
        return self._x_size

    @property
    def y_size(self) -> int:
        return self._y_size

    @property
    def btn_size(self) -> int:
        return 20

    def _setup_ui(self):
        base_layout = QVBoxLayout(self)
        frame = QFrame(self)
        frame.setFrameShadow(QFrame.Raised)
        frame.setFrameShape(QFrame.Box)
        frame.setLineWidth(5)
        frame.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
        base_layout.addWidget(frame)
        sub_layout = QVBoxLayout(frame)
        sub_layout.setContentsMargins(0, 0, 0, 0)
        view = QGraphicsView(frame)
        view.setScene(self._scene)
        sub_layout.addWidget(view)

    # --------------------------------------------------------------------------
    # Qt method overrides
    # --------------------------------------------------------------------------
    def mousePressEvent(self, event: QMouseEvent):
        """Handle mouse press events."""
        if not self._animation_started:
            self._animation_started = True
            self._do_next_update()

    # --------------------------------------------------------------------------
    # Other methods
    # --------------------------------------------------------------------------
    def _do_next_update(self):
        """Perform the next set of cell updates."""
        evt = self._remaining_cell_updates.pop(0)
        self._update_cells({tuple(c): CellContents.from_str(x) for c, x in evt[1]})
        if self._remaining_cell_updates:
            QTimer.singleShot(
                1000 * (self._remaining_cell_updates[0][0] - evt[0]),
                self._do_next_update,
            )

    def _set_cell_image(self, coord: Coord, state: CellContents) -> None:
        """
        Set the image of a cell.

        Arguments:
        coord ((x, y) tuple in grid range)
            The coordinate of the cell.
        state
            The cell_images key for the image to be set.
        """
        if state not in self._cell_images:
            logger.error("Missing cell image for state: %s", state)
            return
        x, y = coord
        b = self._scene.addPixmap(self._cell_images[state])
        b.setPos(x * self.btn_size, y * self.btn_size)

    def _update_cells(self, cell_updates: Mapping[Coord, CellContents]) -> None:
        """
        Called to indicate some cells have changed state.

        :param cell_updates:
            A mapping of cell coordinates to their new state.
        """
        for c, state in cell_updates.items():
            self._set_cell_image(c, state)



# January 2022, Lewis Gaul

__all__ = (
    "RAISED_CELL",
    "SUNKEN_CELL",
    "FlagAction",
    "MinefieldWidgetImplBase",
    "update_cell_images",
)

import abc
import enum
import logging
import os.path
from typing import Dict, Mapping, Optional

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QImage, QPainter, QPixmap
from PyQt5.QtWidgets import QGraphicsScene

from ... import api, paths
from ...core.board import BoardBase
from ...shared.types import CellContents, CellImageType, Coord
from ..state import State


logger = logging.getLogger(__name__)

RAISED_CELL = CellContents.Unclicked
SUNKEN_CELL = CellContents.UnclickedSunken


def update_cell_images(
    cell_images: Dict[CellContents, QPixmap],
    size: int,
    styles: Mapping[CellImageType, str],
    required: CellImageType = CellImageType.ALL,
) -> None:
    """
    Initialise or update the pixmap images for the minefield cells.

    :param cell_images:
        The dictionary to fill with the created pixmap images.
    :param size:
        The size in pixels to make the image (square).
    :param styles:
        The image styles to use.
    :param required:
        Which image types require updating.
    """

    def get_path(subdir: str, style: str, fname: str, *, fallback: bool = True) -> str:
        base_path = paths.IMG_DIR / subdir
        full_path = base_path / style / fname
        if not full_path.exists() and fallback:
            logger.warning(f"Missing image file at {full_path}, using standard style")
            full_path = base_path / "Standard" / fname
        return str(full_path)

    btn_style = styles[CellImageType.BUTTONS]
    mkr_style = styles[CellImageType.MARKERS]
    num_style = styles[CellImageType.NUMBERS]

    if required & CellImageType.BUTTONS:
        cell_images[RAISED_CELL] = _make_pixmap(
            size, get_path("buttons", btn_style, "btn_up.png")
        )
        cell_images[SUNKEN_CELL] = _make_pixmap(
            size, get_path("buttons", btn_style, "btn_down.png")
        )
    if required & (CellImageType.BUTTONS | CellImageType.NUMBERS):
        for i in range(1, 19):
            cell_images[CellContents.Num(i)] = _make_pixmap(
                size,
                get_path("buttons", btn_style, "btn_down.png"),
                get_path("numbers", num_style, f"num{i}.png"),
                propn=7 / 8,
            )
        num0_path = get_path("numbers", num_style, f"num0.png", fallback=False)
        if os.path.exists(num0_path):
            cell_images[CellContents.Num(0)] = _make_pixmap(
                size,
                get_path("buttons", btn_style, "btn_down.png"),
                num0_path,
                propn=7 / 8,
            )
        else:
            cell_images[CellContents.Num(0)] = cell_images[SUNKEN_CELL]
    if required & (CellImageType.BUTTONS | CellImageType.MARKERS):
        for i in range(1, 4):
            cell_images[CellContents.Flag(i)] = _make_pixmap(
                size,
                get_path("buttons", btn_style, "btn_up.png"),
                get_path("markers", mkr_style, f"flag{i}.png"),
                propn=5 / 8,
            )
            cell_images[CellContents.WrongFlag(i)] = _make_pixmap(
                size,
                get_path("buttons", btn_style, "btn_up.png"),
                get_path("markers", mkr_style, f"cross{i}.png"),
                propn=5 / 8,
            )
            cell_images[CellContents.Mine(i)] = _make_pixmap(
                size,
                get_path("buttons", btn_style, "btn_down.png"),
                get_path("markers", mkr_style, f"mine{i}.png"),
                propn=7 / 8,
            )
            cell_images[CellContents.HitMine(i)] = _make_pixmap(
                size,
                get_path("buttons", btn_style, "btn_down_hit.png"),
                get_path("markers", mkr_style, f"mine{i}.png"),
                propn=7 / 8,
            )


def _make_pixmap(
    size: int,
    bg_path: str,
    fg_path: Optional[str] = None,
    *,
    propn: float = 1.0,
) -> QPixmap:
    """
    Create a compound pixmap image, superimposing a foreground over a background.

    :param size:
        The size to create the pixmap.
    :param bg_path:
        Path to background image.
    :param fg_path:
        Path to foreground image, if any.
    :param propn:
        Proportion of the foreground image over the background image.
    :return:
        The created pixmap image.
    """
    if fg_path:
        image = QImage(bg_path).scaled(
            size, size, transformMode=Qt.SmoothTransformation
        )
        fg_size = int(propn * size)
        overlay = QPixmap(fg_path).scaled(
            fg_size, fg_size, transformMode=Qt.SmoothTransformation
        )
        painter = QPainter(image)
        margin = int(size * (1 - propn) / 2)
        painter.drawPixmap(margin, margin, overlay)
        painter.end()
        image = QPixmap.fromImage(image)
    else:
        image = QPixmap(bg_path).scaled(
            size, size, transformMode=Qt.SmoothTransformation
        )
    return image


class FlagAction(enum.Enum):
    FLAG = enum.auto()
    UNFLAG = enum.auto()
    SPLIT = enum.auto()


class MinefieldWidgetImplBase(metaclass=abc.ABCMeta):
    """
    An implementation of the logic that's specific to the given game mode.
    """

    def __init__(
        self, scene: QGraphicsScene, ctrlr: api.AbstractController, state: State
    ):
        self._scene = scene
        self._ctrlr = ctrlr
        self._state = state

    @property
    def _board(self) -> BoardBase:
        return self._ctrlr.board

    @property
    def x_size(self) -> int:
        return self._state.x_size

    @property
    def y_size(self) -> int:
        return self._state.y_size

    @property
    def btn_size(self) -> int:
        return self._state.btn_size

    @abc.abstractmethod
    def set_cell_image(self, coord: Coord, state: CellContents) -> None:
        """Set the image of a cell on the scene."""
        raise NotImplementedError

    @abc.abstractmethod
    def update_cell_images(self, img_type: CellImageType = CellImageType.ALL) -> None:
        """Update the cache of cell images for current size/styles."""
        raise NotImplementedError

    @abc.abstractmethod
    def right_down_action(self, coord: Coord) -> FlagAction:
        """Perform a right click, returning the action that was taken."""
        raise NotImplementedError

    @abc.abstractmethod
    def right_drag_action(self, coord: Coord, action: FlagAction) -> None:
        """Perform a right drag click that matches the given action."""
        raise NotImplementedError



# February 2020, Lewis Gaul

"""
Parse bot messages.

"""

__all__ = ("GENERAL_INFO", "RoomType", "parse_msg")

import argparse
import enum
import logging
import sys
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from minegauler.app.shared.types import Difficulty, GameMode

from . import formatter, utils


logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------
# Message parsing
# ------------------------------------------------------------------------------


class InvalidArgsError(Exception):
    pass


class InvalidMsgError(Exception):
    pass


class PositionalArg:
    def __init__(
        self,
        name: str,
        *,
        parse_name: bool = False,
        nargs: Union[int, str] = 1,
        default: Any = None,
        choices: Optional[Iterable[Any]] = None,
        type: Optional[Callable] = None,
        validate: Optional[Callable] = None,
    ):
        if (isinstance(nargs, int) and nargs < 1) and nargs not in ["?", "*", "+"]:
            raise ValueError(f"Bad nargs value {nargs!r}")
        self.name = name
        self.parse_name = parse_name
        self.nargs = nargs
        if default is None and nargs not in [1, "?"]:
            self.default = []
        else:
            self.default = default
        self._choices = choices
        self._type = type
        self._validate = validate

    def convert(self, value):
        if self._type is not None:
            return self._type(value)
        return value

    def validate(self, value):
        if self._choices is not None and value not in self._choices:
            return False
        if self._validate:
            return self._validate(value)
        return True


class ArgParser(argparse.ArgumentParser):
    """
    A specialised arg parser.

    The list of args to be parsed can contain the following, with no overlap:
     - Ordered positional args at the start
     - Regular argparse args with no order

    Positional args are added with 'add_positional_arg()'. The order of calls to
    this method determines the order the args must appear in. The following
    options are accepted:
     - parse_name: Whether the arg name should be parsed.
     - nargs: The number of args to accept. Accepted values are positive
        integers, "?" for 0 or 1, "*" for 0 or more, "+" for 1 or more.
     - choices: As for 'add_argument()'.
     - type: As for 'add_argument()'.

    Positional args are greedily consumed, but if an arg does not satisfy
    choices or if the 'type' callable raises an exception then no more values
    will be accepted for that arg.

    Examples:
     - If a positional arg has no type/choices restrictions and unbounded
        'nargs' (i.e. set to "*" or "+") then all positional options will be
        consumed by this arg.
     - If a positional arg has unbounded 'nargs' but has a 'type' that raises an
        exception for invalid args, subsequent args will take over the matching.
        Note that if insufficient options are matched (e.g. nargs="+" and no
        options are matched) then parsing ends with a standard error.

    Positional argument parsing ends as soon as an option starting with a dash
    is encountered, or when the positional args are exhausted. The remaining
    options are passed to a standard argparse parser to match args added with
    'add_argument()'.
    """

    def __init__(self):
        super().__init__(add_help=False)
        self._name_parse_args = []
        self._positional_args = []

    def parse_known_args(self, args: Iterable[str], namespace=None):
        """
        Override the default behaviour.
        """
        if namespace is None:
            namespace = argparse.Namespace()

        # Start with positional args.
        args = self._parse_positional_args(args, namespace)
        logger.debug("Remaining after parsing positional args: %s", args)

        # Replace optional args so that they don't require the dashes.
        for i, arg in enumerate(args):
            if arg in self._name_parse_args:
                args[i] = f"--{arg}"

        # Regular parsing of remaining args
        return super().parse_known_args(args, namespace)

    def add_argument(self, name, *args, **kwargs):
        """
        Override the default behaviour.
        """
        if not name.startswith("--"):
            self._name_parse_args.append(name)
            name = "--" + name.lstrip("-")
        super().add_argument(name, *args, **kwargs)

    def error(self, message):
        raise InvalidArgsError(message)

    def add_positional_arg(self, name: str, **kwargs) -> None:
        """
        Add a positional argument for parsing.

        :param name:
            The name of the argument - must be unique.
        :param kwargs:
            Arguments to pass to the 'PositionalArg' class.
        """
        self._positional_args.append(PositionalArg(name, **kwargs))

    def _parse_positional_args(self, kws: Iterable[str], namespace) -> Iterable[str]:
        """
        Parse the positional args.

        :param kws:
            The provided keywords.
        :param namespace:
            The namespace to set argument values in.
        :return:
            The remaining unmatched keywords.
        """
        for arg in self._positional_args:
            logger.debug(f"Parsing positional arg %r, kws: %s", arg.name, kws)
            result, kws = self._parse_single_positional_arg(arg, kws)
            setattr(namespace, arg.name, result)
        return kws

    def _parse_single_positional_arg(
        self, arg: PositionalArg, kws: Iterable[str]
    ) -> Tuple[Any, Iterable[str]]:
        """
        Parse a single positional arg. Raise InvalidArgsError if not enough
        matching args are found.
        """
        required = arg.nargs not in ["?", "*"]
        if isinstance(arg.nargs, int):
            max_matches = arg.nargs
            exp_args_string = str(arg.nargs)
        elif arg.nargs == "?":
            max_matches = 1
            exp_args_string = "optionally one"
        elif arg.nargs == "*":
            max_matches = None
            exp_args_string = "any number of"
        elif arg.nargs == "+":
            max_matches = None
            exp_args_string = "at least one"
        else:
            assert False

        # First parse the arg name if required.
        if kws and arg.parse_name:
            if kws[0] == arg.name:  # Found arg
                kws.pop(0)
            elif not required:  # No match
                return arg.default, kws
            else:
                raise InvalidArgsError(f"Expected to find {arg.name!r}")

        # Now parse argument values.
        matches = []
        while kws and (max_matches is None or len(matches) < max_matches):
            try:
                kw_value = arg.convert(kws[0])
                assert arg.validate(kw_value)
            except Exception as e:
                logger.debug(e)
                if arg.parse_name and not matches:
                    # We parsed the name of the arg, so we expected to find
                    # at least one value...
                    raise InvalidArgsError(
                        f"Got name of positional arg {arg.name!r} but no values"
                    ) from e
                else:
                    break
            else:
                matches.append(kw_value)
                kws.pop(0)

        if required and not matches:
            raise InvalidArgsError(f"Expected {exp_args_string} {arg.name!r} arg")
        elif isinstance(arg.nargs, int) and len(matches) != arg.nargs:
            assert len(matches) < arg.nargs
            raise InvalidArgsError(f"Expected {exp_args_string} {arg.name!r} arg")

        arg_value = arg.default
        if matches:
            if arg.nargs in [1, "?"]:
                assert len(matches) == 1
                arg_value = matches[0]
            else:
                arg_value = matches

        return arg_value, kws


class BotMsgParser(ArgParser):
    def add_game_mode_arg(self):
        self.add_positional_arg(
            "game_mode", nargs="?", type=GameMode.from_str, default=GameMode.REGULAR
        )

    def add_difficulty_arg(self):
        self.add_positional_arg("difficulty", nargs="?", type=Difficulty.from_str)

    def add_per_cell_arg(self):
        self.add_argument("per-cell", type=int, choices=[1, 2, 3])

    def add_drag_select_arg(self):
        def _arg_type(arg):
            if arg == "on":
                return True
            elif arg == "off":
                return False
            else:
                raise InvalidArgsError("Drag select should be one of {'on', 'off'}")

        self.add_argument("drag-select", type=_arg_type)


# ------------------------------------------------------------------------------
# Message handling
# ------------------------------------------------------------------------------


CommandMapping = Dict[Optional[str], Union[Callable, "CommandMapping"]]


class RoomType(enum.Enum):
    GROUP = "group"
    DIRECT = "direct"
    LOCAL = "local"

    def to_cmds(self) -> CommandMapping:
        if self is RoomType.GROUP:
            return _GROUP_COMMANDS
        elif self is RoomType.DIRECT:
            return _DIRECT_COMMANDS
        elif self is RoomType.LOCAL:
            return _LOCAL_COMMANDS
        else:
            assert False


GENERAL_INFO = """\
Welcome to the Minegauler bot!

Instructions for downloading Minegauler can be found in the \
[GitHub repo](https://github.com/LewisGaul/minegauler/blob/main/README.md).

The bot provides the ability to check highscores and matchups for games of \
Minegauler, and also gives notifications whenever anyone in the group sets a \
new highscore.

There are a few settings to filter Minegauler games with:
 - mode: One of 'regular' or 'split-cell' (defaults to 'regular').
 - difficulty: One of 'beginner', 'intermediate', 'expert' or 'master'.  
   By default the combination of beginner, intermediate and expert is used, \
with a score of 1000 used for any difficulties not completed.
 - drag-select: One of 'on' or 'off'.  
   By default no filter is applied (the best time of either mode is used).
 - per-cell: One of 1, 2 or 3.  
   By default no filter is applied (the best time of any mode is used).

All highscores are independent of each of the above modes, and all commands \
accept these filters to view specific times. E.g. 'ranks beginner per-cell 1'.

Commands can be issued in group chat and direct chat. Most of the commands are \
the same, however some are only available in one of the other (e.g. \
'set nickname'). Type 'help' or '?' to check the available commands.

To interact with a bot in Webex Teams group chats, the bot must be tagged. To \
do this, type @ followed by the bot's name. The client should auto-suggest a \
completion, at which point you'll need to press enter or tab to select the \
completion and turn it into a tag.

Some useful common commands:  
'ranks' - display rankings  
'matchups <name> <name> ...' - display comparison of times between users

Some useful group-chat commands:  
'challenge <name> ...' - challenge other users to a game

Some useful direct-chat commands:  
'set nickname' - set your nickname that you use in the Minegauler app
"""


LOCAL_INFO = """\
Welcome to the Minegauler bot!

Instructions for downloading Minegauler can be found in the \
[GitHub repo](https://github.com/LewisGaul/minegauler/blob/main/README.md).

The local bot provides the ability to check highscores and matchups for games \
of Minegauler.

There are a few settings to filter Minegauler games with:
 - mode: One of 'regular' or 'split-cell' (defaults to 'regular').
 - difficulty: One of 'beginner', 'intermediate', 'expert' or 'master'.  
   By default the combination of beginner, intermediate and expert is used, \
with a score of 1000 used for any difficulties not completed.
 - drag-select: One of 'on' or 'off'.  
   By default no filter is applied (the best time of either mode is used).
 - per-cell: One of 1, 2 or 3.  
   By default no filter is applied (the best time of any mode is used).

All highscores are independent of each of the above modes, and all commands \
accept these filters to view specific times. E.g. 'ranks beginner per-cell 1'.

Some useful common commands:  
'ranks' - display rankings  
'player <name>' - display player info  
'matchups <name> <name> ...' - display comparison of times between users
"""


def helpstring(text):
    def decorator(func):
        func.__helpstring__ = text
        return func

    return decorator


def schema(text):
    def decorator(func):
        func.__schema__ = text
        return func

    return decorator


def cmd_help(
    func: Callable, *, only_schema: bool = False, allow_markdown: bool = False
) -> str:
    lines = []
    if not only_schema:
        try:
            lines.append(func.__helpstring__)
        except AttributeError:
            logger.warning(
                "No helpstring found on message handling function %r", func.__name__
            )
    try:
        schema = func.__schema__
    except AttributeError:
        logger.warning("No schema found on message handling function %r", func.__name__)
    else:
        if allow_markdown:
            lines.append(f"`{schema}`")
        else:
            lines.append(schema)

    if not lines:
        return "Unexpected error: unable to get help message\n\n"

    return "\n\n".join(lines)


@helpstring("Get help for a command")
@schema("help [<command>]")
def help_(args, **kwargs):
    allow_markdown = kwargs.get("allow_markdown", False)
    room_type = kwargs.get("room_type", RoomType.DIRECT)
    cmds = room_type.to_cmds()

    linebreak = "\n\n" if allow_markdown else "\n"
    commands = _flatten_cmds(cmds)
    if allow_markdown:
        commands = f"`{commands}`"

    if not args:
        return commands

    func, _ = _map_to_cmd(args, cmds)
    if func is None:
        raise InvalidArgsError(linebreak.join(["Unrecognised command", commands]))
    return cmd_help(func, allow_markdown=allow_markdown)


@helpstring("Get information about the game")
@schema("info")
def info(args, **kwargs):
    # Check no args given.
    BotMsgParser().parse_args(args)
    if kwargs["room_type"] is RoomType.LOCAL:
        return LOCAL_INFO
    else:
        return GENERAL_INFO


@helpstring("Get player info")
@schema(
    "player <name> "
    "[regular | split-cell] "
    "[b[eginner] | i[ntermediate] | e[xpert] | m[aster] | l[udicrous]] "
    "[drag-select {on | off}] "
    "[per-cell {1 | 2 | 3}]"
)
def player(args, username: str, allow_markdown=False, **kwargs):
    parser = BotMsgParser()
    username_choices = list(utils.USER_NAMES)
    if username:
        username_choices.append("me")
    parser.add_positional_arg("username", choices=username_choices)
    parser.add_game_mode_arg()
    parser.add_difficulty_arg()
    parser.add_per_cell_arg()
    parser.add_drag_select_arg()
    args = parser.parse_args(args)
    if args.username == "me":
        args.username = username

    highscores = utils.get_highscores(
        name=utils.USER_NAMES[args.username],
        game_mode=args.game_mode,
        difficulty=args.difficulty,
        drag_select=args.drag_select,
        per_cell=args.per_cell,
    )

    filters_str = formatter.format_filters(
        game_mode=None,
        difficulty=None,
        drag_select=args.drag_select,
        per_cell=args.per_cell,
        no_difficulty=True,
    )
    if filters_str:
        filters_str = " with " + filters_str
    if utils.USER_NAMES[args.username] == args.username:
        username_str = ""
    else:
        username_str = f" ({utils.USER_NAMES[args.username]})"
    lines = [
        "Player info for {name}{username} on {mode} mode{filters}:".format(
            name=args.username,
            username=username_str,
            mode=args.game_mode.value,
            filters=filters_str,
        )
    ]
    lines.extend(
        formatter.format_player_highscores(list(highscores), difficulty=args.difficulty)
    )

    linebreak = "  \n" if allow_markdown else "\n"

    return linebreak.join(lines)


@helpstring("Get rankings")
@schema(
    "ranks "
    "[regular | split-cell] "
    "[b[eginner] | i[ntermediate] | e[xpert] | m[aster] | l[udicrous]] "
    "[drag-select {on | off}] "
    "[per-cell {1 | 2 | 3}]"
)
def ranks(args, **kwargs) -> str:
    allow_markdown = kwargs.get("allow_markdown", False)

    parser = BotMsgParser()
    parser.add_game_mode_arg()
    parser.add_difficulty_arg()
    parser.add_per_cell_arg()
    parser.add_drag_select_arg()
    args = parser.parse_args(args)

    times = utils.get_highscore_times(
        game_mode=args.game_mode,
        difficulty=args.difficulty,
        drag_select=args.drag_select,
        per_cell=args.per_cell,
    )

    lines = [
        "Rankings for {}".format(
            formatter.format_filters(
                game_mode=args.game_mode,
                difficulty=args.difficulty,
                drag_select=args.drag_select,
                per_cell=args.per_cell,
            )
        )
    ]
    ranks = formatter.format_highscore_times(times)
    if allow_markdown:
        ranks = f"```\n{ranks}\n```"
    lines.append(ranks)

    return "\n".join(lines)


@helpstring("Get stats for played games")
@schema(  # @@@ This is bad because it requires knowledge of sub-commands.
    "stats [players ...] "
    "[regular | split-cell] "
    "[b[eginner] | i[ntermediate] | e[xpert] | m[aster] | l[udicrous]] "
    "[drag-select {on | off}] "
    "[per-cell {1 | 2 | 3}]"
)
def stats(args, **kwargs):
    parser = BotMsgParser()
    parser.add_game_mode_arg()
    parser.add_difficulty_arg()
    parser.add_per_cell_arg()
    parser.add_drag_select_arg()
    args = parser.parse_args(args)
    return "Stats"


@helpstring("Get player stats")
@schema("stats players {all | <name> [<name> ...]}")
def stats_players(args, username: str, allow_markdown=False, **kwargs):
    parser = BotMsgParser()
    username_choices = list(utils.USER_NAMES) + ["all"]
    if username:
        username_choices.append("me")
    parser.add_positional_arg("username", nargs="+", choices=username_choices)
    args = parser.parse_args(args)
    if "all" in args.username:
        if len(args.username) > 1:
            raise InvalidArgsError("'all' should be specified without usernames")
        users = utils.USER_NAMES.keys()
    else:
        users = {u if u != "me" else username for u in args.username}

    player_info = [utils.get_player_info(u) for u in users]
    lines = [formatter.format_player_info(player_info)]
    if allow_markdown:
        lines = ["```", *lines, "```"]

    return "\n".join(lines)


@helpstring("Get matchups for given players")
@schema(
    "matchups <name> <name> [<name> ...] "
    "[regular | split-cell] "
    "[b[eginner] | i[ntermediate] | e[xpert] | m[aster] | l[udicrous]] "
    "[drag-select {on | off}] "
    "[per-cell {1 | 2 | 3}]"
)
def matchups(
    args,
    username: str,
    allow_markdown: bool = False,
    room_type=RoomType.DIRECT,
    **kwargs,
):
    parser = BotMsgParser()
    username_choices = list(utils.USER_NAMES)
    if username:
        username_choices.append("me")
    parser.add_positional_arg("username", nargs="+", choices=username_choices)
    parser.add_game_mode_arg()
    parser.add_difficulty_arg()
    parser.add_per_cell_arg()
    parser.add_drag_select_arg()
    args = parser.parse_args(args)
    users = {u if u != "me" else username for u in args.username}
    if len(users) < 2 or len(users) > 5:
        raise InvalidArgsError("Require between 2 and 5 username args")
    names = {utils.USER_NAMES[u] for u in users}

    times = utils.get_highscore_times(
        game_mode=args.game_mode,
        difficulty=args.difficulty,
        drag_select=args.drag_select,
        per_cell=args.per_cell,
        users=names,
    )
    matchups = utils.get_matchups(times)

    if allow_markdown and room_type is RoomType.GROUP:
        users_str = ", ".join(utils.tag_user(u) for u in users)
    else:
        users_str = ", ".join(users)
    lines = [
        "Matchups between {users} for {filters}:".format(
            users=users_str,
            filters=formatter.format_filters(
                game_mode=args.game_mode,
                difficulty=args.difficulty,
                drag_select=args.drag_select,
                per_cell=args.per_cell,
            ),
        )
    ]
    lines.extend(formatter.format_matchups(matchups))
    linebreak = "  \n" if allow_markdown else "\n"

    return linebreak.join(lines)


@helpstring("Get best matchups including at least one of specified players")
@schema(
    "best-matchups [<name> ...] "
    "[regular | split-cell] "
    "[b[eginner] | i[ntermediate] | e[xpert] | m[aster] | l[udicrous]] "
    "[drag-select {on | off}] "
    "[per-cell {1 | 2 | 3}]"
)
def best_matchups(
    args, username: str, allow_markdown=False, room_type=RoomType.DIRECT, **kwargs
):
    parser = BotMsgParser()
    username_choices = list(utils.USER_NAMES)
    if username:
        username_choices.append("me")
    parser.add_positional_arg("username", nargs="*", choices=username_choices)
    parser.add_game_mode_arg()
    parser.add_difficulty_arg()
    parser.add_per_cell_arg()
    parser.add_drag_select_arg()
    args = parser.parse_args(args)
    users = {u if u != "me" else username for u in args.username}
    names = {utils.USER_NAMES[u] for u in users}

    times = utils.get_highscore_times(
        game_mode=args.game_mode,
        difficulty=args.difficulty,
        drag_select=args.drag_select,
        per_cell=args.per_cell,
        users=utils.USER_NAMES.values(),
    )
    matchups = utils.get_matchups(times, include_users=names)[:10]

    if allow_markdown and room_type is RoomType.GROUP:
        users_str = ", ".join(utils.tag_user(u) for u in users)
    else:
        users_str = ", ".join(users)
    if users_str:
        users_str = " including " + users_str
    lines = [
        "Best matchups{users} for {filters}:".format(
            users=users_str,
            filters=formatter.format_filters(
                game_mode=args.game_mode,
                difficulty=args.difficulty,
                drag_select=args.drag_select,
                per_cell=args.per_cell,
            ),
        )
    ]
    lines.extend(formatter.format_matchups(matchups))
    linebreak = "  \n" if allow_markdown else "\n"

    return linebreak.join(lines)


@helpstring("Challenge other players to a game")
@schema(
    "challenge <name> [<name> ...] "
    "[regular | split-cell] "
    "[b[eginner] | i[ntermediate] | e[xpert] | m[aster] | l[udicrous]] "
    "[drag-select {on | off}] "
    "[per-cell {1 | 2 | 3}]"
)
def challenge(args, username: str, **kwargs):
    parser = BotMsgParser()
    parser.add_positional_arg(
        "username",
        nargs="+",
        choices=set(utils.USER_NAMES) - {username} - utils.NO_TAG_USERS,
    )
    parser.add_positional_arg("game_mode", nargs="?", type=GameMode.from_str)
    parser.add_difficulty_arg()
    parser.add_per_cell_arg()
    parser.add_drag_select_arg()
    args = parser.parse_args(args)
    names = {u for u in args.username}

    users_str = ", ".join(utils.tag_user(u) for u in names)
    mode_str = args.game_mode.value + " " if args.game_mode else ""
    mode_str += args.difficulty.name.lower() + " " if args.difficulty else ""
    filters_str = formatter.format_filters(
        game_mode=None,
        difficulty=None,
        drag_select=args.drag_select,
        per_cell=args.per_cell,
        no_difficulty=True,
    )
    if filters_str:
        filters_str = " with " + filters_str

    return "{user} has challenged {opponents} to a {mode}game of Minegauler{filters}".format(
        user=username, opponents=users_str, mode=mode_str, filters=filters_str
    )


@helpstring("Set your nickname")
@schema("set nickname <name>")
def set_nickname(args, username: str, **kwargs):
    new = " ".join(args)
    if len(new) > 10:
        raise InvalidArgsError("Nickname must be no longer than 10 characters")
    for other in utils.USER_NAMES.values():
        if new.lower() == other.lower():
            raise InvalidArgsError(f"Nickname {other!r} already in use")
    if new.lower() in utils.USER_NAMES.keys() and new != username:
        raise InvalidArgsError(f"Cannot set nickname to someone else's username!")
    old = utils.USER_NAMES[username]
    logger.debug("Changing nickname of %s from %r to %r", username, old, new)
    utils.set_user_nickname(username, new)
    return f"Nickname changed from {old!r} to {new!r}"


@helpstring("Add a user to the local rankings")
@schema("user add <name>")
def add_user(args, **kwargs):
    parser = BotMsgParser()
    parser.add_positional_arg("name")
    args = parser.parse_args(args)
    for user in utils.USER_NAMES:
        if args.name.lower() == user.lower():
            raise InvalidMsgError(f"User {user!r} already added")
    logger.debug("Adding user %r", args.name)
    utils.set_user_nickname(args.name, args.name)
    return f"Added user {args.name!r}"


@helpstring("Remove a user from the local rankings")
@schema("user remove <name>")
def remove_user(args, **kwargs):
    parser = BotMsgParser()
    parser.add_positional_arg("name")
    args = parser.parse_args(args)
    try:
        utils.USER_NAMES.pop(args.name)
    except KeyError as e:
        raise InvalidMsgError(f"User {args.name!r} not found") from e
    else:
        utils.save_users_file()
    logger.debug("Removed user %r", args.name)
    return f"Removed user {args.name!r}"


@helpstring("List users on the local rankings")
@schema("user list")
def list_users(args, **kwargs):
    parser = BotMsgParser()
    parser.parse_args(args)
    msg = "The following users are on the local rankings:"
    for user in sorted(utils.USER_NAMES, key=lambda x: x.lower()):
        msg += "\n - " + user
    return msg


_COMMON_COMMANDS = {
    "help": help_,
    "player": player,
    "ranks": ranks,
    "stats": {
        # None: stats,
        "players": stats_players,
    },
    "matchups": matchups,
    "best-matchups": best_matchups,
}

_GROUP_COMMANDS = {
    **_COMMON_COMMANDS,
    "challenge": challenge,
}

_DIRECT_COMMANDS = {
    **_COMMON_COMMANDS,
    "info": info,
    "set": {
        "nickname": set_nickname,
    },
}

_LOCAL_COMMANDS = {
    **_COMMON_COMMANDS,
    "info": info,
    "user": {
        "add": add_user,
        "remove": remove_user,
        "list": list_users,
    },
}


def _map_to_cmd(msg: Iterable[str], cmds: CommandMapping) -> Tuple[Callable, List[str]]:
    func = None
    words = list(msg)

    while words:
        next_word = words[0]
        if next_word in cmds:
            words.pop(0)
            if callable(cmds[next_word]):
                func = cmds[next_word]
                break
            else:
                cmds = cmds[next_word]
        else:
            break
        if None in cmds:
            func = cmds[None]

    return func, words


def _flatten_cmds(cmds: CommandMapping, root: bool = True) -> str:
    to_join = []
    for k, v in cmds.items():
        item = k
        if isinstance(v, dict):
            item += " " + _flatten_cmds(v, root=False)
        if item:
            to_join.append(item)
    ret = " | ".join(to_join)
    if None in cmds:
        ret = f"[{ret}]"
    elif len(cmds) > 1 and not root:
        ret = f"{{{ret}}}"
    return ret


def parse_msg(
    msg: Union[str, List[str]],
    room_type: RoomType,
    *,
    allow_markdown: bool = False,
    **kwargs,
) -> str:
    """
    Parse a message and perform the corresponding action.

    :param msg:
        The message to parse.
    :param room_type:
        The room type the message was received in.
    :param allow_markdown:
        Whether to allow markdown in the response.
    :param kwargs:
        Other arguments to pass on to sub-command functions.
    :return:
        Response text.
    :raise InvalidArgsError:
        Unrecognised command. The text of the error is a suitable error/help
        message.
    """
    orig_msg = msg
    if isinstance(msg, str):
        msg = msg.strip().split()
    else:
        msg = msg.copy()
    if not msg:
        msg = ["help"]
    if msg[-1] in ["?", "help"] and not msg[0] == "help":
        # Change "?" at the end to be treated as help, except for the case where
        # this is a double help intended for a subcommand, e.g.
        # 'help matchups ?'.
        msg.pop(-1)
        msg.insert(0, "help")

    cmds = room_type.to_cmds()
    func = None
    try:
        func, args = _map_to_cmd(msg, cmds)
        if func is None:
            raise InvalidArgsError("Base command not found")
        return func(args, allow_markdown=allow_markdown, room_type=room_type, **kwargs)
    except InvalidMsgError as e:
        logger.debug("Invalid message: %r", orig_msg)
        resp_msg = f"Invalid command: {str(e)}"
        raise InvalidMsgError(resp_msg) from e
    except InvalidArgsError as e:
        logger.debug("Invalid message: %r", orig_msg)
        if func is None:
            raise InvalidArgsError(
                "Unrecognised command - try 'help' or 'info' in direct chat"
            ) from e
        else:
            linebreak = "\n\n" if allow_markdown else "\n"
            resp_msg = cmd_help(func, only_schema=True, allow_markdown=allow_markdown)
            resp_msg = linebreak.join([f"Unrecognised command: {str(e)}", resp_msg])

        raise InvalidArgsError(resp_msg) from e


# ------------------------------------------------------------------------------
# Main
# ------------------------------------------------------------------------------


def main(argv: Optional[List[str]] = None) -> int:
    if argv is None:
        argv = sys.argv[1:]
    rc = 0
    utils.read_users_file()
    try:
        resp = parse_msg(argv, RoomType.LOCAL, username="")
    except (InvalidArgsError, InvalidMsgError) as e:
        resp = str(e)
        rc = 1
    print(resp)
    return rc


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# April 2018, Lewis Gaul

"""
Widgets in the top panel.

Exports
-------
.. class:: PanelWidget
    Widget for the top panel of the Minegauler GUI.

"""

__all__ = ("PanelWidget",)

from typing import Optional, Union

from PyQt5.QtCore import Qt, QTimer, pyqtSignal
from PyQt5.QtGui import QColor, QFont, QPalette, QPixmap
from PyQt5.QtWidgets import QFrame, QHBoxLayout, QLabel, QWidget

from .. import paths
from ..shared.types import FaceState, GameState
from . import state


class PanelWidget(QWidget):
    """
    The panel widget.
    """

    clicked = pyqtSignal()

    def __init__(self, parent: Optional[QWidget], state_: state.State):
        """
        :param parent:
            Parent widget.
        :param state_:
            Shared frontend state.
        """
        super().__init__(parent)
        self._state: state.State = state_

        self.setFixedHeight(40)
        self.setMinimumWidth(120)

        self._mines_counter = _CounterWidget(self, self._state.mines)
        self._face_button = QLabel(self)
        self.timer = Timer(self)
        self.setup_ui()

    def setup_ui(self):
        """
        Set up the widgets contained in the panel.
        """
        layout = QHBoxLayout(self)
        layout.setContentsMargins(6, 2, 6, 2)
        layout.setAlignment(Qt.AlignCenter)
        # Mine counter widget.
        layout.addWidget(self._mines_counter)
        layout.addStretch()
        # Face button.
        self._face_button = QLabel(self)
        self._face_button.setFixedSize(32, 32)
        self._face_button.setFrameShape(QFrame.Panel)
        self._face_button.setFrameShadow(QFrame.Raised)
        self._face_button.setLineWidth(3)
        layout.addWidget(self._face_button)
        self.set_face(FaceState.READY)
        layout.addStretch()
        # Timer widget.
        layout.addWidget(self.timer.widget)

    # --------------------------------------------------------------------------
    # Qt method overrides
    # --------------------------------------------------------------------------
    def mousePressEvent(self, event):
        """Handle mouse press event."""
        super().mousePressEvent(event)
        if event.button() == Qt.LeftButton:
            self._face_button.setFrameShadow(QFrame.Sunken)

    def mouseReleaseEvent(self, event):
        """Handle mouse release event."""
        if event.button() == Qt.LeftButton:
            self._face_button.setFrameShadow(QFrame.Raised)
            if self.rect().contains(event.pos()):
                self.clicked.emit()

    # --------------------------------------------------------------------------
    # Other methods
    # --------------------------------------------------------------------------
    def reset(self) -> None:
        """
        Reset the panel state.
        """
        self.update_game_state(GameState.READY)
        self.set_mines_counter(self._state.mines)
        self.timer.reset()

    def set_face(self, state: Union[FaceState, GameState, str]) -> None:
        """
        Arguments:
        state (str | GameState | FaceState)
        """
        try:
            state = state.value.lower()
        except AttributeError:
            pass
        life = 1
        fname = f"face{life}{state}.png"
        pixmap = QPixmap(str(paths.IMG_DIR / "faces" / fname))
        self._face_button.setPixmap(
            pixmap.scaled(26, 26, transformMode=Qt.SmoothTransformation)
        )

    def at_risk(self) -> None:
        if not self._state.game_status.finished():
            self.set_face(FaceState.ACTIVE)

    def no_risk(self) -> None:
        if not self._state.game_status.finished():
            self.set_face(FaceState.READY)

    def set_mines_counter(self, num: int) -> None:
        """
        This method is to be registered as a callback with a controller, as
        the widget itself can have no way of knowing how many mines are left to
        be found.
        """
        self._mines_counter.set_count(num)

    def update_game_state(self, state: GameState) -> None:
        """
        Receive an update from a backend.

        :param state:
            The new game state.
        """
        if state is GameState.ACTIVE:
            self.timer.start()
        elif state in {GameState.WON, GameState.LOST}:
            self.timer.stop()
            self.set_face(state)
        else:
            self.timer.reset()
            self.set_face(state)


class _CounterWidget(QLabel):
    """A widget template for the counters."""

    def __init__(self, parent: QWidget, count: int = 0):
        super().__init__(parent)
        self.setFrameShadow(QFrame.Sunken)
        self.setFrameShape(QFrame.Panel)
        self.setLineWidth(2)
        self.setFixedSize(39, 26)
        self.setAutoFillBackground(True)
        self.setAlignment(Qt.AlignCenter)
        self._palette = QPalette()
        self._palette.setColor(QPalette.Window, QColor("black"))
        self._palette.setColor(QPalette.WindowText, QColor("red"))
        self.setPalette(self._palette)
        self._font = QFont()
        self._font.setFamily("Helvetica")
        self._font.setPixelSize(13)
        self._font.setBold(True)
        self.setFont(self._font)
        self._count = count
        self.set_count(self._count)

    def set_count(self, count: int) -> None:
        """Set the value of the counter."""
        self.setText(f"{min(999, abs(count)):03d}")
        if count < 0:
            self._invert(False)
        else:
            self._invert(True)
        self._count = count

    def _invert(self, normal: bool = True) -> None:
        """Invert the colours."""
        if normal:
            self._palette.setColor(QPalette.Window, QColor("black"))
            self._palette.setColor(QPalette.WindowText, QColor("red"))
        else:
            self._palette.setColor(QPalette.Window, QColor("red"))
            self._palette.setColor(QPalette.WindowText, QColor("black"))
        self.setPalette(self._palette)


class Timer(QTimer):
    """A timer for the panel."""

    def __init__(self, parent):
        super().__init__()
        self.widget = _CounterWidget(parent)
        self.timeout.connect(self.update)
        self.seconds = 0

    # -------
    # Methods from parent class
    # -------
    def start(self):
        self.seconds = 1
        self.set_time(self.seconds)
        super().start(1000)  # Update every second

    def update(self):
        self.seconds += 1
        self.set_time(self.seconds)

    # -------
    # Public methods
    # -------
    def reset(self) -> None:
        """
        Reset the timer, stopping it if necessary.
        """
        self.stop()
        self.set_time(0)

    def set_time(self, seconds):
        self.widget.setText("{:03d}".format(min(seconds, 999)))



# January 2022, Lewis Gaul

__all__ = ("MinefieldWidgetImpl",)

import logging
from typing import Dict

from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QGraphicsScene

from ... import api
from ...core.split_cell import Coord
from ...shared.types import CellContents, CellImageType
from ..state import State
from . import _base
from ._base import FlagAction, update_cell_images


logger = logging.getLogger(__name__)


class MinefieldWidgetImpl(_base.MinefieldWidgetImplBase):
    """
    Minefield widget for regular game.
    """

    def __init__(
        self, scene: QGraphicsScene, ctrlr: api.AbstractController, state: State
    ):
        super().__init__(scene, ctrlr, state)
        self._small_cell_images: Dict[CellContents, QPixmap] = {}
        self._large_cell_images: Dict[CellContents, QPixmap] = {}
        self.update_cell_images()

    def set_cell_image(self, coord: Coord, state: CellContents) -> None:
        """Set the image of a cell on the scene."""
        try:
            if coord.is_split:
                img = self._small_cell_images[state]
            else:
                img = self._large_cell_images[state]
        except KeyError:
            logger.error("Missing cell image for state: %s", state)
            return
        b = self._scene.addPixmap(img)
        b.setPos(coord.x * self.btn_size, coord.y * self.btn_size)

    def update_cell_images(self, img_type: CellImageType = CellImageType.ALL) -> None:
        update_cell_images(
            self._small_cell_images, self.btn_size, self._state.styles, img_type
        )
        update_cell_images(
            self._large_cell_images, self.btn_size * 2, self._state.styles, img_type
        )

    def right_down_action(self, coord: Coord) -> FlagAction:
        """Perform a right click, returning the action that was taken."""
        if coord.is_split or self._ctrlr.board[coord] is not CellContents.Unclicked:
            self._ctrlr.flag_cell(coord)
            if self._ctrlr.board[coord] is CellContents.Unclicked:
                return FlagAction.UNFLAG
            else:
                return FlagAction.FLAG
        else:
            self._ctrlr.split_cell(coord)
            return FlagAction.SPLIT

    def right_drag_action(self, coord: Coord, action: FlagAction) -> None:
        """Perform a right drag click that matches the given action."""
        if action is FlagAction.FLAG:
            self._ctrlr.flag_cell(coord, flag_only=True)
        elif action is FlagAction.UNFLAG:
            self._ctrlr.remove_cell_flags(coord)
        elif action is FlagAction.SPLIT:
            self._ctrlr.split_cell(coord)
        else:
            assert False, f"Unexpected action {action!r}"



# October 2021, Lewis Gaul

__all__ = ("Game",)

import logging
import time
from typing import Iterable, Mapping

from ...shared.types import CellContents, Difficulty, GameMode, GameState
from ..game import GameBase, GameNotStartedError, _check_coord, _ignore_if_not
from .board import Board
from .minefield import Minefield
from .types import Coord


logger = logging.getLogger(__name__)


class Game(GameBase):
    """A split-cells minesweeper game."""

    mode = GameMode.REGULAR
    minefield_cls = Minefield
    board_cls = Board

    mf: Minefield
    board: Board

    _diff_pairs = [
        # fmt: off
        (Difficulty.BEGINNER,     ( 8,  8,   5)),
        (Difficulty.INTERMEDIATE, (16, 16,  20)),
        (Difficulty.EXPERT,       (30, 16,  50)),
        (Difficulty.MASTER,       (30, 30, 100)),
        (Difficulty.LUDICROUS,    (50, 50, 400)),
        # fmt: on
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Used for tracking numbers that should be revealed for performance.
        self._revealed_board = Board(self.x_size, self.y_size)

    # ---------------------
    # Abstract methods
    # ---------------------
    def _make_board(self) -> Board:
        return Board(self.x_size, self.y_size)

    def get_rem_3bv(self) -> int:
        if self.state is GameState.READY:
            try:
                return self.mf.bbbv
            except AttributeError:
                raise GameNotStartedError("Minefield not yet created") from None
        elif self.state is GameState.WON:
            return 0

        # Partially completed board - do the real work!

        # TODO: This is too slow!

        partial_mf = Minefield.from_coords(
            self.mf.all_coords,
            mine_coords=self.mf.mine_coords,
            per_cell=self.per_cell,
        )
        # Replace any openings already found with normal clicks (ones).
        for c in self.board.all_coords:
            if type(self.board[c]) is CellContents.Num:
                partial_mf.completed_board[c] = CellContents.Num(1)
        # Find the openings which remain.
        rem_opening_coords = {c for opening in partial_mf.openings for c in opening}
        # Count the number of essential clicks that have already been
        # done by counting clicked cells minus the ones at the edge of
        # an undiscovered opening.
        num_coords = {
            c for c in self.board.all_coords if type(self.board[c]) is CellContents.Num
        }
        completed_3bv = len(num_coords - rem_opening_coords)
        # Add necessary splits that have already been done.
        big_cells_correctly_split = {
            c.get_big_cell_coord()
            for c in self.board.all_coords
            if c.is_split and c not in self.mf.mine_coords
        }
        completed_3bv += len(big_cells_correctly_split)
        return partial_mf._calc_3bv() - completed_3bv

    def _populate_minefield(self, coord: Coord) -> None:
        """Create the minefield in response to a cell being selected."""
        if self.first_success:
            safe_coords = [
                small
                for big in self.board.get_nbrs(coord, include_origin=True)
                for small in big.split()
            ]
            logger.debug(
                "Trying to create minefield with the following safe coordinates: %s",
                safe_coords,
            )
            try:
                self.mf.populate(safe_coords)
            except ValueError:
                try:
                    logger.info(
                        "Unable to give opening on the first click, "
                        "trying to ensure a safe click"
                    )
                    self.mf.populate(safe_coords=coord.split())
                except ValueError:
                    logger.info("Unable to give even a single safe cell")
                    self.mf.populate()
            else:
                logger.debug("Minefield populated")
        else:
            logger.debug("Creating minefield without guaranteed first click success")
            self.mf.populate()

    def _is_complete(self) -> bool:
        return all(
            type(self.board[c]) is CellContents.Num
            or all(small in self.mf.mine_coords for small in c.get_small_cell_coords())
            for c in self.board.all_coords
        )

    def _select_cell_action(self, coord: Coord) -> None:
        """Implementation of the action of selecting/clicking a cell."""
        if self._revealed_board[coord] is CellContents.Unclicked:
            # TODO: This is a bit of a hacky place to do this...
            self._revealed_board = self._calc_revealed_board()

        if coord.is_split:
            if coord in self.mf.mine_coords:
                logger.debug("Mine hit at %s", coord)
                self._set_cell(coord, CellContents.HitMine(self.mf[coord]))
                self._finalise_lost_game()
            else:
                logger.debug("Regular cell revealed")
                self._set_cell(coord, self._revealed_board[coord])
        else:
            small_cells = coord.split()
            if any(c in self.mf.mine_coords for c in small_cells):
                logger.debug("Mine hit in large cell containing %s", coord)
                for c in small_cells:
                    if self.mf[c] > 0:
                        self._set_cell(c, CellContents.HitMine(self.mf[c]))
                self._finalise_lost_game()
            else:
                assert type(self._revealed_board[coord]) is CellContents.Num
                cell_num = self._revealed_board[coord].num
                if cell_num > 0:  # regular num revealed
                    logger.debug("Regular cell revealed")
                    self._set_cell(coord, self._revealed_board[coord])
                else:  # opening hit
                    logger.debug("Opening found")
                    # Use pre-calculated 'revealed board' for performance.
                    checked = set()
                    blank_nbrs = {coord}
                    all_nbrs = {coord}
                    while blank_nbrs - checked:
                        c = (blank_nbrs - checked).pop()
                        logger.debug("Checking neighbours of %s", c)
                        cur_nbrs = set(
                            c
                            for c in self.board.get_nbrs(c)
                            if self.board[c] is CellContents.Unclicked
                        )
                        all_nbrs |= cur_nbrs
                        blank_nbrs |= {
                            c
                            for c in cur_nbrs
                            if self._revealed_board[c] == CellContents.Num(0)
                        }
                        checked.add(c)
                    logger.debug("Opening cells deduced")
                    for c in all_nbrs:
                        self._set_cell(c, self._revealed_board[c])

    # ---------------------
    # Other methods
    # ---------------------
    def _calc_revealed_board(self) -> Board:
        board = Board(self.x_size, self.y_size)
        for coord in board.all_coords:
            mines = sum(self.mf[small] for small in coord.get_small_cell_coords())
            if mines > 0:
                board[coord] = CellContents.Flag(mines)
            else:
                num = sum(
                    self.mf[small]
                    for nbr in board.get_nbrs(coord)
                    for small in nbr.get_small_cell_coords()
                )
                board[coord] = CellContents.Num(num)
        return board

    def _calc_nbr_mines(self, coord: Coord) -> int:
        return sum(
            self.mf[c]
            for nbr in self.board.get_nbrs(coord)
            for c in nbr.get_small_cell_coords()
        )

    def _update_board_numbers(self, coords: Iterable[Coord]) -> None:
        """
        Calculate the numbers contained in the board given the split cell
        situation.
        """
        for c in coords:
            num = CellContents.Num(self._calc_nbr_mines(c))
            if type(self.board[c]) is CellContents.Num:
                self._set_cell(c, num)
            self._revealed_board[c] = num

    def _finalise_lost_game(self) -> None:
        logger.info("Game lost")
        self.end_time = time.time()
        self.state = GameState.LOST
        for c in self.mf.mine_coords:
            if self.board[self.board.get_coord_at(c.x, c.y)] is CellContents.Unclicked:
                self._set_cell(c, CellContents.Mine(self.mf[c]))
        for c in self.board.all_coords:
            if (
                type(self.board[c]) is CellContents.Flag
                and self.board[c].num != self.mf[c]
            ):
                self._set_cell(c, CellContents.WrongFlag(self.board[c].num))

    @_check_coord
    @_ignore_if_not(
        game_state=[GameState.READY, GameState.ACTIVE],
        cell_state=CellContents.Unclicked,
    )
    def split_cell(self, coord: Coord) -> Mapping[Coord, CellContents]:
        small_cells = coord.split()

        just_started = False
        if self.state is GameState.READY:
            if not self.mf.populated:
                logger.debug(
                    "Creating minefield without guaranteed first click success"
                )
                self.mf.populate()
            self.state = GameState.ACTIVE
            self.start_time = time.time()
            just_started = True
            self._revealed_board = self._calc_revealed_board()

        if not any(c in self.mf.mine_coords for c in small_cells):
            logger.debug("Incorrect cell split %s", coord)
            self._set_cell(coord, CellContents.WrongFlag(1))
            self._finalise_lost_game()
            if self.state.finished() and just_started:
                self.end_time = self.start_time
        else:
            logger.debug("Splitting cell %s", coord)
            nbrs = self.board.get_nbrs(coord)
            self.board.split_coord(coord)
            self._revealed_board.split_coord(coord)
            self._cell_updates.update({c: CellContents.Unclicked for c in small_cells})
            self._update_board_numbers((*nbrs, *small_cells))
        try:
            return self._cell_updates
        finally:
            self._cell_updates = {}



# December 2019, Felix Gaul

__all__ = ("SQLiteDB",)

import logging
import os.path
import pathlib
import sqlite3
from typing import Iterable, Optional, Tuple

import attr

from ..shared.types import Difficulty, GameMode, PathLike
from .base import AbstractHighscoresDB, HighscoreStruct, SQLMixin


logger = logging.getLogger(__name__)


class SQLiteDB(SQLMixin, AbstractHighscoresDB):
    """Database of local highscores."""

    def __init__(self, path: PathLike):
        self._path = pathlib.Path(path)
        if self._path.is_file():
            self._conn = sqlite3.connect(str(self._path))
        else:
            logger.debug("Creating SQLite highscores DB")
            os.makedirs(self._path.parent, exist_ok=True)
            self._conn = sqlite3.connect(str(self._path))

            for t in self.TABLES.values():
                self.execute(self._get_create_table_sql(t), commit=True)
            self.execute("PRAGMA user_version = 1")

    @property
    def conn(self) -> sqlite3.Connection:
        return self._conn

    @property
    def path(self) -> pathlib.Path:
        return self._path

    def get_db_version(self) -> int:
        """Get the database version number."""
        cursor = self.execute("PRAGMA user_version")
        return self.extract_single_elem(cursor)

    def get_highscores(
        self,
        *,
        game_mode: Optional[GameMode] = None,
        difficulty: Optional[Difficulty] = None,
        per_cell: Optional[int] = None,
        drag_select: Optional[bool] = None,
        name: Optional[str] = None,
    ) -> Iterable[HighscoreStruct]:
        super().get_highscores(
            game_mode=game_mode,
            difficulty=difficulty,
            per_cell=per_cell,
            drag_select=drag_select,
            name=name,
        )
        if game_mode is None:
            modes = list(GameMode)
        else:
            modes = [game_mode]
        ret = []
        for mode in modes:
            self._conn.row_factory = lambda cursor, row: HighscoreStruct(
                game_mode=mode,
                **{col[0]: row[i] for i, col in enumerate(cursor.description)},
            )
            cursor = self.execute(
                self._get_select_highscores_sql(
                    game_mode=mode,
                    difficulty=difficulty,
                    per_cell=per_cell,
                    drag_select=drag_select,
                    name=name,
                )
            )
            ret.extend(cursor.fetchall())
        self._conn.row_factory = None
        return ret

    def count_highscores(self) -> int:
        """Count the number of rows in the highscores table."""
        super().count_highscores()
        return sum(
            self.extract_single_elem(
                self.execute(self._get_highscores_count_sql(game_mode=mode))
            )
            for mode in GameMode
        )

    def insert_highscores(self, highscores: Iterable[HighscoreStruct]) -> int:
        super().insert_highscores(highscores)
        orig_count = self.count_highscores()
        for mode in GameMode:
            mode_rows = [attr.astuple(h)[1:] for h in highscores if h.game_mode is mode]
            self.executemany(
                self._get_insert_highscore_sql(fmt="?", game_mode=mode),
                mode_rows,
                commit=True,
            )
        return self.count_highscores() - orig_count

    def execute(
        self, cmd: str, params: Tuple = (), *, commit=False, **cursor_args
    ) -> sqlite3.Cursor:
        return super().execute(cmd, params, commit=commit, **cursor_args)

    def executemany(
        self, cmd: str, params: Iterable[Tuple] = (), *, commit=False, **cursor_args
    ) -> sqlite3.Cursor:
        return super().executemany(cmd, params, commit=commit, **cursor_args)



# February 2022, Lewis Gaul

"""
Highscores handling.

"""

__all__ = (
    "HighscoreReadError",
    "HighscoreSettingsStruct",
    "HighscoreStruct",
    "SQLiteDB",
    "filter_and_sort",
    "get_highscores",
    "insert_highscore",
    "is_highscore_new_best",
)

import logging
import threading
from typing import Dict, Iterable, List, Optional

import attr
import requests

from .. import paths
from .._version import __version__
from ..shared import utils
from ..shared.types import Difficulty, GameMode, PathLike
from . import compat
from .base import AbstractHighscoresDB, HighscoreSettingsStruct, HighscoreStruct
from .compat import HighscoreReadError

# from .mysql import MySQLDB  # Do not uncomment this without adding dependency on mysql connector
from .sqlite import SQLiteDB


logger = logging.getLogger(__name__)

_REMOTE_POST_URL = "http://minegauler.lewisgaul.co.uk/api/v1/highscore"

_default_local_db = SQLiteDB(paths.HIGHSCORES_FILE)


def get_highscores(
    *,
    database: Optional[AbstractHighscoresDB] = None,
    settings: Optional[HighscoreSettingsStruct] = None,
    game_mode: Optional[GameMode] = None,
    difficulty: Optional[Difficulty] = None,
    per_cell: Optional[int] = None,
    drag_select: Optional[bool] = None,
    name: Optional[str] = None,
) -> Iterable[HighscoreStruct]:
    """
    Fetch highscores from a database.

    :param database:
        The database to fetch from, defaults to the local highscores DB.
    :param settings:
        Optionally specify settings to filter by.
    :param game_mode:
        Optionally specify game mode to filter by. Ignored if settings given.
    :param difficulty:
        Optionally specify difficulty to filter by. Ignored if settings given.
    :param per_cell:
        Optionally specify per-cell to filter by. Ignored if settings given.
    :param drag_select:
        Optionally specify drag-select to filter by. Ignored if settings given.
    :param name:
        Optionally specify a name to filter by.
    """
    if database is None:
        database = _default_local_db
    if settings is not None:
        game_mode = settings.game_mode
        difficulty = settings.difficulty
        per_cell = settings.per_cell
        drag_select = settings.drag_select
    return database.get_highscores(
        game_mode=game_mode,
        difficulty=difficulty,
        per_cell=per_cell,
        drag_select=drag_select,
        name=name,
    )


def insert_highscore(
    highscore: HighscoreStruct,
    *,
    database: Optional[AbstractHighscoresDB] = None,
    post_remote: bool = True,
) -> None:
    """
    Insert a highscore into a database.

    :param highscore:
        The highscore to insert.
    :param database:
        The database to insert into, defaults to the local highscores DB.
    :param post_remote:
        Whether to post the highscore to the remote master DB.
    """
    if database is None:
        database = _default_local_db
    database.insert_highscores([highscore])

    if post_remote:

        def _post_catch_exc():
            try:
                _post_highscore_to_remote(highscore)
            except Exception:
                logger.exception("Failed to insert highscore into remote DB")

        threading.Thread(target=_post_catch_exc).start()


def retrieve_highscores(path: PathLike) -> int:
    """
    Insert highscores from another DB at the given path.

    Handles reading in older highscore formats.

    :param path:
        Path to highscores, may be file or directory as required.
    :return:
        Number of inserted highscores.
    :raise HighscoreReadError:
        If unable to read highscores from the given path.
    """
    return _default_local_db.insert_highscores(compat.read_highscores(path))


def filter_and_sort(
    highscores: Iterable[HighscoreStruct],
    sort_key: str = "time",
    filters: Dict[str, Optional[str]] = {},
) -> List[HighscoreStruct]:
    """
    Filter and sort an iterable of highscores.

    :param highscores:
        The iterable of highscores to filter and sort.
    :param sort_key:
        What to sort by.
    :param filters:
        What filters to apply.
    :return:
        A new list of highscores.
    """
    # TODO: Generalise/tidy up...
    ret = []
    filters = {k: f for k, f in filters.items() if f}
    for hs in highscores:
        all_pass = True
        if "flagging" in filters:
            if (
                filters["flagging"] == "F"
                and not utils.is_flagging_threshold(hs.flagging)
                or filters["flagging"] == "NF"
                and utils.is_flagging_threshold(hs.flagging)
            ):
                all_pass = False
        if "name" in filters and filters["name"].lower() != hs.name.lower():
            all_pass = False
        if all_pass:
            # All filters satisfied.
            ret.append(hs)
    # Sort first by either time or 3bv/s, then by 3bv if there's a tie
    #  (higher 3bv higher for equal time, lower for equal 3bv/s).
    if sort_key == "time":
        ret.sort(key=lambda h: (h.elapsed, -h.bbbv))
    elif sort_key == "3bv/s":
        ret.sort(key=lambda h: (h.bbbvps, -h.bbbv), reverse=True)
    if "name" not in filters:
        # If no name filter, only include best highscore for each name.
        names = []
        i = 0
        while i < len(ret):
            hs = ret[i]
            name = hs.name.lower()
            if name in names:
                ret.pop(i)
            else:
                names.append(name)
                i += 1
    return ret


def is_highscore_new_best(
    highscore: HighscoreStruct, all_highscores: Iterable[HighscoreStruct]
) -> Optional[str]:
    """
    Test to see if a new top highscore has been set.

    :param highscore:
        The highscore to check.
    :param all_highscores:
        The list of highscores to check against. May or may not include the
        highscore being checked.
    :return:
        If a new highscore was set, return which category it was set in. If not,
        return None.
    """
    all_highscores = list(all_highscores)
    top_time = filter_and_sort(all_highscores, "time", {"name": highscore.name})
    top_3bvps = filter_and_sort(all_highscores, "3bv/s", {"name": highscore.name})
    if not top_time or highscore.elapsed <= top_time[0].elapsed:
        return "time"
    elif not top_3bvps or highscore.bbbvps >= top_3bvps[0].bbbvps:
        return "3bv/s"
    else:
        return None


def _post_highscore_to_remote(highscore: HighscoreStruct):
    """Send a highscore to the remote server to be added to the remote DB."""
    logger.info("Posting highscore to remote")
    requests.post(
        _REMOTE_POST_URL,
        json={
            "highscore": attr.asdict(highscore),
            "app_version": __version__,
        },
        timeout=5,
    )



# October 2021, Lewis Gaul

__all__ = ("Game",)

import logging
import time

from ...shared.types import CellContents, Difficulty, GameMode, GameState
from ..game import GameBase, GameNotStartedError
from .board import Board
from .minefield import Minefield
from .types import Coord


logger = logging.getLogger(__name__)


class Game(GameBase):
    """A regular minesweeper game."""

    mode = GameMode.REGULAR
    minefield_cls = Minefield
    board_cls = Board

    mf: Minefield
    board: Board

    _diff_pairs = [
        # fmt: off
        (Difficulty.BEGINNER,     ( 8,  8,  10)),
        (Difficulty.INTERMEDIATE, (16, 16,  40)),
        (Difficulty.EXPERT,       (30, 16,  99)),
        (Difficulty.MASTER,       (30, 30, 200)),
        (Difficulty.LUDICROUS,    (50, 50, 625)),
        # fmt: on
    ]

    # ---------------------
    # Abstract methods
    # ---------------------
    def _make_board(self) -> Board:
        return Board(self.x_size, self.y_size)

    def get_rem_3bv(self) -> int:
        if self.state is GameState.READY:
            try:
                return self.mf.bbbv
            except AttributeError:
                raise GameNotStartedError("Minefield not yet created") from None
        elif self.state is GameState.WON:
            return 0

        # Partially completed board - do the real work!

        partial_mf = Minefield.from_coords(
            self.mf.all_coords,
            mine_coords=self.mf.mine_coords,
            per_cell=self.per_cell,
        )
        # Replace any openings already found with normal clicks (ones).
        for c in self.board.all_coords:
            if type(self.board[c]) is CellContents.Num:
                partial_mf.completed_board[c] = CellContents.Num(1)
        # Find the openings which remain.
        rem_opening_coords = {c for opening in partial_mf.openings for c in opening}
        # Count the number of essential clicks that have already been
        # done by counting clicked cells minus the ones at the edge of
        # an undiscovered opening.
        completed_3bv = len(
            {
                c
                for c in self.board.all_coords
                if type(self.board[c]) is CellContents.Num
            }
            - rem_opening_coords
        )
        return partial_mf._calc_3bv() - completed_3bv

    def _populate_minefield(self, coord: Coord) -> None:
        """Create the minefield in response to a cell being selected."""
        if self.first_success:
            safe_coords = self.board.get_nbrs(coord, include_origin=True)
            logger.debug(
                "Trying to create minefield with the following safe coordinates: %s",
                safe_coords,
            )
            try:
                self.mf.populate(safe_coords)
            except ValueError:
                logger.info(
                    "Unable to give opening on the first click, "
                    "still ensuring a safe click"
                )
                # This should be guaranteed to succeed.
                self.mf.populate(safe_coords=[coord])
            else:
                logger.debug("Successfully created minefield")
        else:
            logger.debug("Creating minefield without guaranteed first click success")
            self.mf.populate()

    def _is_complete(self) -> bool:
        return all(
            type(self.board[c]) is CellContents.Num or c in self.mf.mine_coords
            for c in self.board.all_coords
        )

    def _select_cell_action(self, coord: Coord) -> None:
        """
        Implementation of the action of selecting/clicking a cell.
        """
        if coord in self.mf.mine_coords:
            logger.debug("Mine hit at %s", coord)
            self._set_cell(coord, CellContents.HitMine(self.mf[coord]))
            self.lives_remaining -= 1

            if self.lives_remaining == 0:
                logger.info("Game lost")
                self.end_time = time.time()
                self.state = GameState.LOST

                for c in self.mf.all_coords:
                    if (
                        c in self.mf.mine_coords
                        and self.board[c] is CellContents.Unclicked
                    ):
                        self._set_cell(c, CellContents.Mine(self.mf[c]))

                    elif (
                        type(self.board[c]) is CellContents.Flag
                        and self.board[c] != self.mf.completed_board[c]
                    ):
                        self._set_cell(c, CellContents.WrongFlag(self.board[c].num))
            else:
                self.mines_remaining -= self.mf[coord]
        elif self.mf.completed_board[coord] is CellContents.Num(0):
            for full_opening in self.mf.openings:
                if coord in full_opening:
                    # Found the opening, quit the loop here.
                    logger.debug("Opening hit: %s", full_opening)
                    break
            else:
                raise RuntimeError(f"Coordinate {coord} not found in openings")

            # Get the propagation of cells forming part of the opening.
            opening = set()  # Coords belonging to the opening
            check = {coord}  # Coords whose neighbours need checking
            while check:
                c = check.pop()
                unclicked_nbrs = {
                    z
                    for z in self.board.get_nbrs(c, include_origin=True)
                    if self.board[z] is CellContents.Unclicked
                }
                check |= {
                    z
                    for z in unclicked_nbrs - opening
                    if self.mf.completed_board[z] is CellContents.Num(0)
                }
                opening |= unclicked_nbrs

            logger.debug("Propagated opening: %s", list(opening))
            for c in opening:
                self._set_cell(c, self.mf.completed_board[c])
        else:
            logger.debug("Regular cell revealed")
            self._set_cell(coord, self.mf.completed_board[coord])



# December 2018, Lewis Gaul

"""
Entry point for the application.

"""

import logging
import sys

from . import core, frontend, paths, shared


logger = logging.getLogger(__name__)

ctrlr = None
gui = None


def setup_logging():
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    # Create file handler which logs debug messages.
    fh = logging.FileHandler("runtime.log")
    fh.setLevel(logging.DEBUG)
    # Create console handler with a higher log level.
    ch = logging.StreamHandler()
    ch.setLevel(logging.WARNING)
    # Create formatter and add it to the handlers.
    formatter = logging.Formatter("%(asctime)s[%(levelname)s](%(name)s) %(message)s")
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    # Add the handlers to the logger.
    root.addHandler(fh)
    root.addHandler(ch)


def main() -> int:
    global ctrlr, gui

    setup_logging()

    read_settings = shared.read_settings_from_file(paths.SETTINGS_FILE)

    if read_settings:
        game_opts = shared.GameOptsStruct.from_structs(read_settings)
        gui_opts = shared.GUIOptsStruct.from_structs(read_settings)
        logger.info("Settings read from file")
    else:
        logger.info("Using default settings")
        game_opts = shared.GameOptsStruct()
        gui_opts = shared.GUIOptsStruct()
    logger.debug("Game options: %s", game_opts)
    logger.debug("GUI options: %s", gui_opts)

    logger.info("Starting up")

    # Create core controller.
    ctrlr = core.UberController(game_opts)
    # Init frontend and create controller.
    frontend.init_app()
    frontend_state = frontend.state.State.from_opts(game_opts, gui_opts)
    frontend_state.difficulty = ctrlr.get_game_info().difficulty
    gui = frontend.MinegaulerGUI(ctrlr, frontend_state)
    # Register frontend with core controller.
    ctrlr.register_listener(gui)

    # Run the app.
    logger.debug("Entering event loop")
    rc = frontend.run_app(gui)
    logger.debug("Exiting event loop")

    persist_settings = shared.AllOptsStruct.from_structs(
        ctrlr.get_game_options(), gui.get_gui_opts()
    )
    shared.write_settings_to_file(persist_settings, paths.SETTINGS_FILE)

    logger.info("Exiting with exit code %d", rc)

    return rc


if __name__ == "__main__":
    sys.exit(main())



# October 2021, Lewis Gaul

__all__ = ("Coord", "Minefield", "Board", "Game", "CreateController", "GameController")

from .board import Board
from .controller import CreateController, GameController
from .game import Game
from .minefield import Minefield
from .types import Coord



# February 2022, Lewis Gaul

"""
Compatibility with v1 SQLite highscore format.

This version adds split-cell mode highscore support in v4.1.2 minegauler.

"""

__all__ = ("read_highscores",)

import sqlite3
from typing import Iterable

from ...shared.types import PathLike
from ..base import HighscoreStruct


_TABLE_NAMES = ["regular", "split_cell"]


def read_highscores(path: PathLike) -> Iterable[HighscoreStruct]:
    ret = set()
    with sqlite3.connect(str(path)) as conn:
        for table_name in _TABLE_NAMES:
            cursor = conn.execute(f"SELECT * FROM {table_name}")
            for row in cursor:
                ret.add(
                    HighscoreStruct(
                        game_mode=table_name,
                        difficulty=row[0],
                        per_cell=row[1],
                        drag_select=row[2],
                        name=row[3],
                        timestamp=row[4],
                        elapsed=row[5],
                        bbbv=row[6],
                        bbbvps=row[7],
                        flagging=row[8],
                    )
                )
    return ret



# October 2021, Lewis Gaul

__all__ = ("ControllerBase", "GameControllerBase", "CreateControllerBase", "SharedInfo")

import abc
import json
import logging
import os.path
from os import PathLike
from typing import Dict, Optional, Type

import attr

from ..shared import GameOptsStruct
from ..shared.types import CellContents, Coord, Difficulty, GameMode, GameState
from . import api
from .board import BoardBase
from .game import GameBase
from .minefield import MinefieldBase


logger = logging.getLogger(__name__)


@attr.attrs(auto_attribs=True, kw_only=True)
class SharedInfo:
    """
    Information to pass to frontends.

    Elements:
    cell_updates
        Dictionary of updates to cells, mapping the coordinate to the new
        contents of the cell.
    game_state
        The state of the game.
    mines_remaining
        The number of mines remaining to be found, given by
        [total mines] - [number of flags].
        Can be negative if there are too many flags.
    lives_remaining
        The number of lives remaining.
    """

    cell_updates: Optional[Dict[Coord, CellContents]] = None
    game_state: GameState = GameState.READY
    mines_remaining: int = 0
    lives_remaining: int = 0


class ControllerBase(api.AbstractController, metaclass=abc.ABCMeta):
    """Base controller class, generic over game mode."""

    # Remove abstractmethods.
    switch_game_mode = None
    switch_ui_mode = None
    reset_settings = None

    mode: GameMode
    minefield_cls: Type[MinefieldBase]
    board_cls: Type[BoardBase]
    game_cls: Type[GameBase]

    def __init__(
        self,
        opts: GameOptsStruct,
        *,
        notif: api.AbstractListener,
    ):
        """
        :param opts:
            Game options.
        :param notif:
            A notifier defining callbacks.
        """
        super().__init__(opts)
        # Use a reference to the given opts rather than a copy.
        self._opts = opts
        self._notif = notif

    @property
    @abc.abstractmethod
    def difficulty(self) -> Difficulty:
        raise NotImplementedError

    def set_difficulty(self, difficulty: Difficulty) -> None:
        x, y, m = self.game_cls.difficulty_to_values(difficulty)
        self.resize_board(x, y, m)

    def resize_board(self, x_size: int, y_size: int, mines: int) -> None:
        """See AbstractController."""
        super().resize_board(x_size=x_size, y_size=y_size, mines=mines)
        if (
            x_size == self._opts.x_size
            and y_size == self._opts.y_size
            and mines == self._opts.mines
        ):
            logger.info(
                "No resize required as the parameters are unchanged, starting a new game"
            )
            self.new_game()
        else:
            logger.info(
                "Resizing board from %sx%s with %s mines to %sx%s with %s mines",
                self._opts.x_size,
                self._opts.y_size,
                self._opts.mines,
                x_size,
                y_size,
                mines,
            )
            self._opts.x_size = x_size
            self._opts.y_size = y_size
            self._opts.mines = mines
            self._notif.resize_minefield(x_size, y_size)
            self._notif.set_mines(mines)
            self.new_game()
            self._notif.set_difficulty(
                self.game_cls.difficulty_from_values(x_size, y_size, mines)
            )

    @staticmethod
    def _save_minefield(mf: MinefieldBase, file: PathLike) -> None:
        """
        Save a minefield to file.

        :param mf:
            The minefield to save.
        :param file:
            The path of the file to save at.
        :raises OSError:
            If saving to file fails.
        """
        if os.path.isfile(file):
            logger.warning("Overwriting file at %s", file)
        with open(file, mode="w", encoding="utf-8") as f:
            json.dump(mf.to_json(), f)


class GameControllerBase(ControllerBase, metaclass=abc.ABCMeta):
    """Base game controller class, generic over game mode."""

    def __init__(
        self,
        opts: GameOptsStruct,
        *,
        notif: api.AbstractListener,
    ):
        """
        :param opts:
            Game options.
        :param notif:
            A notifier defining callbacks.
        """
        super().__init__(opts, notif=notif)
        self.game = self.game_cls(
            x_size=self._opts.x_size,
            y_size=self._opts.y_size,
            mines=self._opts.mines,
            per_cell=self._opts.per_cell,
            lives=self._opts.lives,
            first_success=self._opts.first_success,
        )
        self._last_update = SharedInfo()
        # self._send_updates()
        self._notif.set_mines(self._opts.mines)

    @property
    def difficulty(self) -> Difficulty:
        return self.game.difficulty

    @property
    def board(self) -> BoardBase:
        return self.game.board

    def get_game_info(self) -> api.GameInfo:
        """Get info about the current game."""
        ret = api.GameInfo(
            game_state=self.game.state,
            x_size=self.game.x_size,
            y_size=self.game.y_size,
            mines=self.game.mines,
            difficulty=self.game.difficulty,
            per_cell=self.game.per_cell,
            first_success=self.game.first_success,
            mode=self.mode,
            minefield_known=self.game.minefield_known,
        )
        if self.game.state.started():
            ret.started_info = api.GameInfo.StartedInfo(
                start_time=self.game.start_time,
                elapsed=self.game.get_elapsed(),
                bbbv=self.game.mf.bbbv,
                rem_bbbv=self.game.get_rem_3bv(),
                bbbvps=self.game.get_3bvps(),
                prop_complete=self.game.get_prop_complete(),
                prop_flagging=self.game.get_flag_proportion(),
            )
        return ret

    # --------------------------------------------------------------------------
    # Methods triggered by user interaction
    # --------------------------------------------------------------------------
    def new_game(self) -> None:
        """See AbstractController."""
        super().new_game()
        if self._opts.mines > self._opts.per_cell * (
            self._opts.x_size * self._opts.y_size - 1
        ):
            # This is needed since it's possible to create a board with more
            # mines than is normally allowed.
            logger.debug(
                "Reducing number of mines from %d to %d because they don't fit",
                self._opts.mines,
                self._opts.x_size * self._opts.y_size - 1,
            )
            self._opts.mines = self._opts.x_size * self._opts.y_size - 1
            self._notif.set_mines(self._opts.mines)
        self.game = self.game_cls(
            x_size=self._opts.x_size,
            y_size=self._opts.y_size,
            mines=self._opts.mines,
            per_cell=self._opts.per_cell,
            lives=self._opts.lives,
            first_success=self._opts.first_success,
        )
        self._send_reset_update()

    def restart_game(self) -> None:
        """See AbstractController."""
        if not self.game.mf.populated:
            return
        super().restart_game()
        self.game = self.game_cls.from_minefield(
            self.game.mf,
            lives=self._opts.lives,
        )
        self._send_reset_update()

    def select_cell(self, coord: Coord) -> None:
        """See AbstractController."""
        super().select_cell(coord)
        cells = self.game.select_cell(coord)
        self._send_updates(cells)

    def chord_on_cell(self, coord: Coord) -> None:
        """See AbstractController."""
        super().chord_on_cell(coord)
        cells = self.game.chord_on_cell(coord)
        self._send_updates(cells)

    def remove_cell_flags(self, coord: Coord) -> None:
        """See AbstractController."""
        super().remove_cell_flags(coord)
        self.game.set_cell_flags(coord, 0)
        self._send_updates({coord: self.board[coord]})

    def set_first_success(self, value: bool) -> None:
        """
        Set whether the first click should be a guaranteed success.
        """
        super().set_first_success(value)
        self._opts.first_success = value
        if not self.game.state.started():
            self.game.first_success = value

    def set_per_cell(self, value: int) -> None:
        """
        Set the maximum number of mines per cell.
        """
        super().set_per_cell(value)
        if value < 1:
            raise ValueError(
                f"Max number of mines per cell must be at least 1, got {value}"
            )
        self._opts.per_cell = value
        # If the game is not started and the minefield is not known then the
        # new per-cell value should be picked up immediately, and the board
        # cleared of any flags (e.g. 3-flag cells may no longer be allowed!).
        if not (self.game.state.started() or self.game.minefield_known):
            self.new_game()

    def save_current_minefield(self, file: PathLike) -> None:
        """
        Save the current minefield to file.

        :param file:
            The location of the file to save to. Should have the extension
            ".mgb".
        :raises RuntimeError:
            If the game is not finished.
        :raises OSError:
            If saving to file fails.
        """
        super().save_current_minefield(file)
        if not self.game.state.finished():
            raise RuntimeError("Can only save minefields when the game is completed")
        self._save_minefield(self.game.mf, file)

    def load_minefield(self, file: PathLike) -> None:
        """Load a minefield from file."""
        with open(file, mode="r", encoding="utf-8") as f:
            mf = self.minefield_cls.from_json(json.load(f))

        logger.debug(
            "Loaded minefield from file (%d x %d, %d mines)",
            mf.x_size,
            mf.y_size,
            mf.mines,
        )
        self.resize_board(mf.x_size, mf.y_size, mf.mines)
        self.game = self.game_cls.from_minefield(mf, lives=self._opts.lives)

    # --------------------------------------------------------------------------
    # Helper methods
    # --------------------------------------------------------------------------
    def _send_updates(
        self, cells_updated: Optional[Dict[Coord, CellContents]] = None
    ) -> None:
        """Send updates to registered listeners."""
        update = SharedInfo(
            cell_updates=cells_updated,
            mines_remaining=self.game.mines_remaining,
            lives_remaining=self.game.lives_remaining,
            game_state=self.game.state,
        )

        # Send updates to registered listeners.
        if update.cell_updates:
            self._notif.update_cells(update.cell_updates)
        if update.mines_remaining != self._last_update.mines_remaining:
            self._notif.update_mines_remaining(update.mines_remaining)
        # if update.lives_remaining != self._last_update.lives_remaining:
        #     self._notif.update_lives_remaining(update.lives_remaining)
        if update.game_state is not self._last_update.game_state:
            self._notif.update_game_state(update.game_state)

        self._last_update = update

    def _send_reset_update(self) -> None:
        """Send an update to reset the board."""
        self._notif.reset()
        self._send_updates()


class CreateControllerBase(ControllerBase, metaclass=abc.ABCMeta):
    """Base create controller class, generic over game mode."""

    # Remove abstractmethod - always handled by game controller.
    load_minefield = None

    def __init__(
        self,
        opts: GameOptsStruct,
        *,
        notif: api.AbstractListener,
    ):
        """
        :param opts:
            Game options.
        :param notif:
            A notifier defining callbacks.
        """
        super().__init__(opts, notif=notif)
        self._board: BoardBase = self._make_board()
        self._flags: int = 0
        self._notif.set_mines(self._flags)

    @abc.abstractmethod
    def _make_board(self) -> BoardBase:
        raise NotImplementedError

    @property
    def board(self) -> BoardBase:
        return self._board

    @property
    def difficulty(self) -> Difficulty:
        return self.game_cls.difficulty_from_values(
            self._opts.x_size, self._opts.y_size, self._flags
        )

    def get_game_info(self) -> api.GameInfo:
        return api.GameInfo(
            game_state=GameState.ACTIVE,
            x_size=self._opts.x_size,
            y_size=self._opts.y_size,
            mines=self._flags,
            difficulty=self.difficulty,
            per_cell=self._opts.per_cell,
            first_success=self._opts.first_success,
            mode=self.mode,
            minefield_known=True,
        )

    def new_game(self) -> None:
        """See AbstractController."""
        super().new_game()
        self._board = self._make_board()
        self._flags = 0
        self._notif.reset()

    def restart_game(self) -> None:
        """See AbstractController."""
        super().restart_game()
        self.new_game()

    def chord_on_cell(self, coord: Coord) -> None:
        super().chord_on_cell(coord)

    def remove_cell_flags(self, coord: Coord) -> None:
        super().remove_cell_flags(coord)
        cell = self.board[coord]
        if not isinstance(cell, CellContents.Mine):
            return
        self.board[coord] = CellContents.Unclicked
        self._flags -= cell.num
        self._notif.update_cells({coord: self.board[coord]})
        self._notif.update_mines_remaining(self._flags)

    def set_first_success(self, value: bool) -> None:
        super().set_first_success(value)
        self._opts.first_success = value

    def set_per_cell(self, value: int) -> None:
        super().set_per_cell(value)
        self._opts.per_cell = value

    def save_current_minefield(self, file: PathLike) -> None:
        """
        Save the current created minefield to file.

        :param file:
            The location of the file to save to. Should have the extension
            ".mgb".
        :raises OSError:
            If saving to file fails.
        """
        super().save_current_minefield(file)
        coords = []
        for c in self.board.all_coords:
            if type(self.board[c]) is CellContents.Mine:
                coords.extend([c] * self.board[c].num)
        mf = self.minefield_cls.from_coords(
            self.board.all_underlying_coords,
            mine_coords=coords,
            per_cell=self._opts.per_cell,
        )
        self._save_minefield(mf, file)



# October 2021, Lewis Gaul

__all__ = ("Coord",)

from typing import Iterable

from ...shared.types import Coord as CoordBase


class Coord(CoordBase):
    fields = ("x", "y", "is_split")
    __slots__ = fields

    def __init__(self, x: int, y: int, is_split: bool):
        """
        :param x:
            The x coord of the underlying small cell (top-left for large cells).
        :param y:
            The y coord of the underlying small cell (top-left for large cells).
        :param is_split:
            Whether the coord corresponds to a small cell.
        """
        if not is_split and (x % 2 or y % 2):
            raise ValueError("Unsplit coords must have even values of x and y")
        self.x = x
        self.y = y
        self.is_split = is_split

    def get_small_cell_coords(self) -> Iterable["Coord"]:
        if self.is_split:
            return (self,)
        else:
            return (
                Coord(self.x, self.y, True),
                Coord(self.x + 1, self.y, True),
                Coord(self.x, self.y + 1, True),
                Coord(self.x + 1, self.y + 1, True),
            )

    def get_big_cell_coord(self) -> "Coord":
        if self.is_split:
            return Coord(self.x // 2 * 2, self.y // 2 * 2, False)
        else:
            return self

    def split(self) -> Iterable["Coord"]:
        if self.is_split:
            raise TypeError(f"Not able to split coord {(self.x, self.y)}")
        return self.get_small_cell_coords()



# October 2021, Lewis Gaul

__all__ = ("Minefield", "RegularMinefieldBase")

import abc
from typing import Any, Iterable, List, Mapping, Optional, TypeVar

from ...shared import utils
from ...shared.types import CellContents
from ...shared.types import Coord as CoordBase
from ..board import BoardBase
from ..minefield import MinefieldBase
from .board import Board
from .types import Coord


C = TypeVar("C", bound=CoordBase)
B = TypeVar("B", bound=BoardBase)


class RegularMinefieldBase(MinefieldBase[C, B], metaclass=abc.ABCMeta):
    """The base for a minefield with regular coords."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.x_size: int = max({c.x for c in self.all_coords}) + 1
        self.y_size: int = max({c.y for c in self.all_coords}) + 1

    def __repr__(self):
        return f"<{self.x_size}x{self.y_size} minefield with {self.mines} mines>"

    def __str__(self):
        max_nr_mines = max(self[c] for c in self.all_coords)
        cell_size = len(repr(max_nr_mines))

        cell = "{:>%d}" % cell_size
        ret = ""
        for y in range(self.y_size):
            for x in range(self.x_size):
                ret += cell.format(self[Coord(x, y)]) + " "
            ret = ret[:-1]  # Remove trailing space
            ret += "\n"
        ret = ret[:-1]  # Remove trailing newline

        return ret

    @classmethod
    def from_dimensions(
        cls, x_size: int, y_size: int, *, mines: int, per_cell: int = 1
    ) -> "RegularMinefieldBase":
        """
        :param x_size:
            Number of columns in the grid.
        :param y_size:
            Number of rows in the grid.
        :param mines:
            The number of mines to randomly place.
        :param per_cell:
            Maximum number of mines per cell.
        :raise ValueError:
            If the number of mines is too high.
        """
        all_coords = [Coord(x, y) for x in range(x_size) for y in range(y_size)]
        return cls(all_coords, mines=mines, per_cell=per_cell)

    @classmethod
    def from_grid(
        cls, grid: utils.Grid, *, per_cell: int = 1
    ) -> "RegularMinefieldBase":
        """
        :param grid:
            A `Grid` instance containing int number of mines in each cell.
        :param per_cell:
            Maximum number of mines per cell.
        :raise ValueError:
            If any of the number of mines is too high.
        """
        mine_coords = []
        for c in grid.all_coords:
            for _ in range(grid[c]):
                mine_coords.append(Coord(*c))
        return cls.from_coords(
            (Coord(*c) for c in grid.all_coords),
            mine_coords=mine_coords,
            per_cell=per_cell,
        )

    @classmethod
    def from_2d_array(cls, array, *, per_cell: int = 1) -> "RegularMinefieldBase":
        """
        :param array:
            2D array containing int number of mines in each cell.
        :param per_cell:
            Maximum number of mines per cell.
        :raise ValueError:
            If any of the number of mines is too high.
        """
        return cls.from_grid(utils.Grid.from_2d_array(array), per_cell=per_cell)


class Minefield(RegularMinefieldBase[Coord, Board]):
    """A regular minesweeper minefield."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._openings: Optional[List[List[Coord]]] = None

    @classmethod
    def from_json(cls, obj: Mapping[str, Any]) -> "Minefield":
        """
        Create a minefield instance from a JSON encoding.

        :param obj:
            The dictionary obtained from decoding JSON. Must contain the
            following fields: 'x_size', 'y_size', 'mine_coords'.
        :raise ValueError:
            If the dictionary is missing required fields.
        """
        try:
            return cls.from_coords(
                (
                    Coord(x, y)
                    for x in range(obj["x_size"])
                    for y in range(obj["y_size"])
                ),
                mine_coords=[Coord(*c) for c in obj["mine_coords"]],
                per_cell=obj.get("per_cell", 1),
            )
        except KeyError as e:
            raise ValueError(
                "Missing key in dictionary when trying to create minefield"
            ) from e

    def to_json(self) -> Mapping[str, Any]:
        return dict(
            type="regular",
            x_size=self.x_size,
            y_size=self.y_size,
            mine_coords=self.mine_coords,
            per_cell=self.per_cell,
        )

    def _get_nbrs(self, coord: Coord, *, include_origin=False) -> Iterable[Coord]:
        """Get coordinates of neighbouring cells."""
        x, y = coord
        nbrs = []
        for i in range(max(0, x - 1), min(self.x_size, x + 2)):
            for j in range(max(0, y - 1), min(self.y_size, y + 2)):
                nbrs.append(Coord(i, j))
        if not include_origin:
            nbrs.remove(coord)
        return nbrs

    def _calc_3bv(self) -> int:
        """Calculate the 3bv of the board."""
        clicks = len(self.openings)
        exposed = len({c for opening in self.openings for c in opening})
        clicks += self.x_size * self.y_size - len(set(self.mine_coords)) - exposed
        return clicks

    def _calc_completed_board(self) -> Board:
        """
        Create the completed board with the flags and numbers that should be
        seen upon game completion.
        """
        completed_board = Board(self.x_size, self.y_size)
        completed_board.fill(CellContents.Num(0))
        for c in self.all_coords:
            mines = self[c]
            if mines > 0:
                completed_board[c] = CellContents.Flag(mines)
                for nbr in self._get_nbrs(c):
                    # For neighbouring cells that don't contain mines, increment
                    # their number.
                    if nbr not in self.mine_coords:
                        completed_board[nbr] += mines
        return completed_board

    def _find_openings(self) -> List[List[Coord]]:
        """
        Find the openings of the board.

        A list of openings is stored, each represented as a list of
        coordinates belonging to that opening. Note that each cell
        cannot belong to multiple openings.
        """
        openings = []
        blanks_to_check = {
            c for c in self.all_coords if self.completed_board[c] is CellContents.Num(0)
        }
        while blanks_to_check:
            orig_coord = blanks_to_check.pop()
            # If the coordinate is part of an opening and hasn't already been
            # considered, start a new opening.
            opening = {orig_coord}  # Coords belonging to the opening
            check = {orig_coord}  # Coords whose neighbours need checking
            while check:
                coord = check.pop()
                nbrs = set(self._get_nbrs(coord))
                check |= {
                    c
                    for c in nbrs - opening
                    if self.completed_board[c] is CellContents.Num(0)
                }
                opening |= nbrs
            openings.append(sorted(opening))
            blanks_to_check -= opening
        return openings



# October 2021, Lewis Gaul

__all__ = ("Coord",)

from typing import Tuple

from ...shared.types import Coord as CoordBase


class Coord(Tuple[int, int], CoordBase):
    def __new__(cls, x: int, y: int):
        self = super().__new__(cls, (x, y))
        self.x = x
        self.y = y
        return self



# March 2018, Lewis Gaul

"""
General utilities.

Exports
-------
.. class:: AllOptsStruct
    A structure class containing all persisted options.

.. class:: GUIOptsStruct
    A structure class containing persisted GUI options.

.. class:: GameOptsStruct
    A structure class containing persisted game options.

.. class:: Grid
    Representation of a 2D array.

.. class:: StructConstructorMixin
    A mixin for structure classes.

.. function:: format_timestamp
    Format a timestamp.

.. function:: is_flagging_threshold
    Check whether flagging threshold is met.

.. function:: read_settings_from_file
    Read persisted settings.

.. function:: write_settings_to_file
    Persist settings to file.

"""

__all__ = (
    "AllOptsStruct",
    "GUIOptsStruct",
    "GameOptsStruct",
    "Grid",
    "StructConstructorMixin",
    "format_timestamp",
    "is_flagging_threshold",
    "read_settings_from_file",
    "write_settings_to_file",
)

import json
import logging
import os
import time
from typing import Any, Dict, Iterable, List, Mapping, Optional, Tuple

import attr

from .types import CellImageType, GameMode, PathLike


logger = logging.getLogger(__name__)


class Grid(list):
    """
    Grid representation using a list of lists (2D array).

    Attributes:
    x_size (int > 0)
        The number of columns.
    y_size (int > 0)
        The number of rows.
    all_coords ([(int, int), ...])
        List of all coordinates in the grid.
    """

    def __init__(self, x_size: int, y_size: int, *, fill: Any = 0):
        """
        Arguments:
        x_size (int > 0)
            The number of columns.
        y_size (int > 0)
            The number of rows.
        fill=0 (object)
            What to fill the grid with.
        """
        super().__init__()
        for _ in range(y_size):
            row = x_size * [fill]
            self.append(row)
        self.x_size: int = x_size
        self.y_size: int = y_size
        self.all_coords: List[Tuple[int, int]] = [
            (x, y) for x in range(x_size) for y in range(y_size)
        ]

    def __repr__(self):
        return f"<{self.x_size}x{self.y_size} grid>"

    def __str__(self, mapping=None, cell_size=None):
        """
        Convert the grid to a string in an aligned format. The __repr__ method
        is used to display the objects inside the grid unless the mapping
        argument is given.

        Arguments:
        mapping=None (dict | callable | None)
            A mapping to apply to all objects contained within the grid. The
            result of the mapping will be converted to a string and displayed.
            If a mapping is specified, a cell size should also be given.
        cell_size=None (int | None)
            The size to display a grid cell as. Defaults to the maximum size of
            the representation of all the objects contained in the grid.
        """
        # @@@LG Some attention please :)

        # Use max length of object representation if no cell size given.
        if cell_size is None:
            cell_size = max([len(repr(obj)) for row in self for obj in row])

        cell = f"{{:>{cell_size}}}"
        ret = ""
        for row in self:
            for obj in row:
                if isinstance(mapping, dict):
                    rep = str(mapping[obj]) if obj in mapping else repr(obj)
                elif mapping is not None:
                    rep = str(mapping(obj))
                else:
                    rep = repr(obj)
                ret += cell.format(rep[:cell_size]) + " "
            ret = ret[:-1]  # Remove trailing space
            ret += "\n"
        ret = ret[:-1]  # Remove trailing newline

        return ret

    def __getitem__(self, key):
        if isinstance(key, tuple) and len(key) == 2:
            x, y = key
            return super().__getitem__(y)[x]
        else:
            raise TypeError("Grid keys should be tuple coordinates of the form (0, 1)")

    def __setitem__(self, key, value):
        if isinstance(key, tuple) and len(key) == 2:
            x, y = key
            super().__getitem__(y)[x] = value
        else:
            raise TypeError("Grid keys should be tuple coordinates of the form (0, 1)")

    @classmethod
    def from_2d_array(cls, array):
        """
        Create an instance using a 2-dimensional array.

        Arguments:
        array ([[object, ...], ...])
            The array to use in creating the grid instance.

        Return: Grid
            The resulting grid.
        """
        x_size = len(array[0])
        y_size = len(array)
        grid = cls(x_size, y_size)
        for coord in grid.all_coords:
            x, y = coord
            grid[coord] = array[y][x]
        return grid

    def fill(self, item):
        """
        Fill the grid with a given object.

        Arguments:
        item (object)
            The item to fill the grid with.
        """
        for row in self:
            for i in range(len(row)):
                row[i] = item

    def get_nbrs(
        self, coord: Tuple[int, int], *, include_origin=False
    ) -> Iterable[Tuple[int, int]]:
        """
        Get a list of the coordinates of neighbouring cells.

        Arguments:
        coord ((int, int), within grid boundaries)
            The coordinate to check.
        include_origin=False (bool)
            Whether to include the original coordinate, coord, in the list.

        Return: [(int, int), ...]
            List of coordinates within the boundaries of the grid.
        """
        x, y = coord
        nbrs = []
        for i in range(max(0, x - 1), min(self.x_size, x + 2)):
            for j in range(max(0, y - 1), min(self.y_size, y + 2)):
                nbrs.append((i, j))
        if not include_origin:
            nbrs.remove(coord)
        return nbrs

    def copy(self):
        ret = Grid(self.x_size, self.y_size)
        for coord in self.all_coords:
            ret[coord] = self[coord]
        return ret

    def is_coord_in_grid(self, coord: Tuple[int, int]) -> bool:
        x, y = coord
        return 0 <= x < self.x_size and 0 <= y < self.y_size


class StructConstructorMixin:
    """
    A mixin class adding methods for ways to create instances.
    """

    @classmethod
    def from_structs(cls, *structs):
        """
        Create an instance using namespace(s) containing the required fields.

        Later arguments take precedence.
        """
        dict_ = {}
        for struct in structs:
            dict_.update(attr.asdict(struct))
        return cls.from_dict(dict_)

    @classmethod
    def from_dict(cls, dict_: Dict[str, Any]):
        """
        Create an instance from a dictionary.

        Ignores extra attributes.
        """
        args = {a: v for a, v in dict_.items() if a in attr.fields_dict(cls)}
        return cls(**args)

    def copy(self):
        """
        Create and return a copy of the instance.

        This is a shallow copy.
        """
        return self.from_structs(self)


@attr.attrs(auto_attribs=True)
class GameOptsStruct(StructConstructorMixin):
    """
    Structure of game options.
    """

    x_size: int = 8
    y_size: int = 8
    mines: int = 10
    first_success: bool = True
    per_cell: int = 1
    lives: int = 1
    mode: GameMode = GameMode.REGULAR


@attr.attrs(auto_attribs=True)
class GUIOptsStruct(StructConstructorMixin):
    """
    Structure of GUI options.
    """

    btn_size: int = 16
    drag_select: bool = False
    name: str = ""
    styles: Mapping[CellImageType, str] = {
        CellImageType.BUTTONS: "Standard",
        CellImageType.NUMBERS: "Standard",
        CellImageType.MARKERS: "Standard",
    }


@attr.attrs(auto_attribs=True)
class AllOptsStruct(GameOptsStruct, GUIOptsStruct):
    """
    Structure containing all application options.
    """

    def encode_to_json(self) -> Dict[str, Any]:
        ret = attr.asdict(self)
        ret["styles"] = {k.name: v for k, v in self.styles.items()}
        ret["mode"] = ret["mode"].name
        return ret

    @classmethod
    def decode_from_json(cls, dict_: Dict[str, Any]) -> "AllOptsStruct":
        dict_["styles"] = {
            getattr(CellImageType, k): v for k, v in dict_["styles"].items()
        }
        dict_["mode"] = getattr(GameMode, dict_.get("mode", "REGULAR"))
        return cls(**dict_)


def is_flagging_threshold(proportion: float) -> bool:
    """Does the given proportion correspond to a board solved with 'flagging'?"""
    return proportion > 0.1


def read_settings_from_file(file: os.PathLike) -> Optional[AllOptsStruct]:
    logger.info("Reading settings from file: %s", file)

    read_settings = None
    try:
        with open(file, mode="r", encoding="utf-8") as f:
            read_settings = AllOptsStruct.decode_from_json(json.load(f))
    except FileNotFoundError:
        logger.info("Settings file not found")
    except json.JSONDecodeError:
        logger.warning("Unable to decode settings from file")
    except Exception:
        logger.warning("Unexpected error reading settings from file", exc_info=True)

    return read_settings


def write_settings_to_file(settings: AllOptsStruct, file: PathLike) -> None:
    logger.info("Saving settings to file: %s", file)
    logger.debug("%s", settings)
    try:
        with open(file, mode="w", encoding="utf-8") as f:
            json.dump(settings.encode_to_json(), f, indent=2)
    except Exception:
        logger.exception("Unexpected error writing settings to file")


def format_timestamp(timestamp: float) -> str:
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(timestamp))



# October 2021, Lewis Gaul

__all__ = ("GameController", "CreateController")

import logging

from ...shared.types import CellContents, GameMode
from ..controller import CreateControllerBase, GameControllerBase
from .board import Board
from .game import Game
from .minefield import Minefield
from .types import Coord


logger = logging.getLogger(__name__)


class _ControllerMixin:
    """Mixin for the controller classes."""

    mode = GameMode.REGULAR
    minefield_cls = Minefield
    board_cls = Board
    game_cls = Game

    @property
    def board(self) -> Board:
        return super().board


class GameController(_ControllerMixin, GameControllerBase):
    """GameController for a regular minesweeper game."""

    def flag_cell(self, coord: Coord, *, flag_only: bool = False) -> None:
        """See AbstractController."""
        super().flag_cell(coord, flag_only=flag_only)

        cell_state = self.board[coord]
        if cell_state is CellContents.Unclicked:
            self.game.set_cell_flags(coord, 1)
        elif isinstance(cell_state, CellContents.Flag):
            if cell_state.num >= self.game.per_cell:
                if flag_only:
                    return
                self.game.set_cell_flags(coord, 0)
            else:
                self.game.set_cell_flags(coord, cell_state.num + 1)

        self._send_updates({coord: self.board[coord]})


class CreateController(_ControllerMixin, CreateControllerBase):
    """A controller for creating boards."""

    def _make_board(self) -> Board:
        return Board(self._opts.x_size, self._opts.y_size)

    def select_cell(self, coord: Coord) -> None:
        super().select_cell(coord)
        if self.board[coord] is CellContents.Unclicked:
            self.board[coord] = CellContents.Num(0)
        elif isinstance(self.board[coord], CellContents.Num):
            self.board[coord] += 1
        else:
            return
        self._notif.update_cells({coord: self.board[coord]})

    def flag_cell(self, coord: Coord, *, flag_only: bool = False) -> None:
        super().flag_cell(coord)
        cell = self.board[coord]

        if cell is CellContents.Unclicked:
            self.board[coord] = CellContents.Mine(1)
            self._flags += 1
        elif isinstance(cell, CellContents.Mine):
            if cell.num == self._opts.per_cell:
                if flag_only:
                    return
                self.board[coord] = CellContents.Unclicked
                self._flags -= self._opts.per_cell
            else:
                self.board[coord] += 1
                self._flags += 1
        elif isinstance(cell, CellContents.Num):
            self.board[coord] = CellContents.Unclicked
        else:
            return
        self._notif.update_cells({coord: self.board[coord]})
        self._notif.update_mines_remaining(self._flags)



# April 2018, Lewis Gaul

__all__ = ("MinefieldWidget",)

import functools
import logging
import time
from typing import Callable, Iterable, List, Mapping, Optional, Set, Type

from PyQt5.QtCore import QSize, Qt, pyqtSignal
from PyQt5.QtGui import QMouseEvent
from PyQt5.QtWidgets import (
    QApplication,
    QGraphicsScene,
    QGraphicsView,
    QSizePolicy,
    QWidget,
)

from ...core import BoardBase, api
from ...shared.types import CellContents, CellImageType, Coord, GameMode
from ..state import State
from ..utils import CellUpdate_T, MouseMove
from . import regular, simulate, split_cell
from ._base import RAISED_CELL, SUNKEN_CELL, FlagAction, MinefieldWidgetImplBase


logger = logging.getLogger(__name__)


_IMPLS: Mapping[GameMode, Type[MinefieldWidgetImplBase]] = {
    GameMode.REGULAR: regular.MinefieldWidgetImpl,
    GameMode.SPLIT_CELL: split_cell.MinefieldWidgetImpl,
}


def _filter_left_and_right(mouse_event_func: Callable):
    """
    Decorator for mouse event methods to filter out buttons that aren't the
    standard left or right mouse buttons.
    """

    @functools.wraps(mouse_event_func)
    def wrapper(self, event: QMouseEvent):
        if event.button() not in [Qt.LeftButton, Qt.RightButton, Qt.NoButton]:
            return
        event = QMouseEvent(
            event.type(),
            event.localPos(),
            event.windowPos(),
            event.screenPos(),
            event.button(),
            event.buttons() & (Qt.LeftButton | Qt.RightButton),
            event.modifiers(),
            event.source(),
        )
        return mouse_event_func(self, event)

    return wrapper


class MinefieldWidget(QGraphicsView):
    """
    The minefield widget.
    """

    at_risk_signal = pyqtSignal()
    no_risk_signal = pyqtSignal()
    size_changed = pyqtSignal()

    def __init__(
        self,
        parent: Optional[QWidget],
        ctrlr: api.AbstractController,
        state: State,
    ):
        super().__init__(parent)
        logger.info("Initialising minefield widget")
        self._ctrlr: api.AbstractController = ctrlr
        self._state: State = state

        self._scene = QGraphicsScene()
        self.setScene(self._scene)
        self.setStyleSheet("border: 0px")
        self.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
        self.setMaximumSize(self.sizeHint())

        self._impl: MinefieldWidgetImplBase = _IMPLS[state.game_mode](
            self._scene, ctrlr, state
        )

        # Keep track of mouse button states.
        self._mouse_coord: Optional[Coord] = None
        self._both_mouse_buttons_pressed: bool = False
        self._await_release_all_buttons: bool = False
        self._was_double_left_click: bool = False
        self._right_drag_action: FlagAction = FlagAction.FLAG

        # Set of coords for cells which are sunken.
        self._sunken_cells: Set[Coord] = set()

        # Mouse tracking info, for simulating a played game.
        self._mouse_tracking: List[MouseMove] = []
        self._mouse_events: List[CellUpdate_T] = []
        self._first_click_time: Optional[int] = None

        self.reset()

    @property
    def _board(self) -> BoardBase:
        return self._ctrlr.board

    @property
    def _elapsed(self) -> float:
        if self._first_click_time:
            return time.time() - self._first_click_time
        else:
            self._first_click_time = time.time()
            return 0

    @property
    def x_size(self) -> int:
        return self._state.x_size

    @property
    def y_size(self) -> int:
        return self._state.y_size

    @property
    def btn_size(self) -> int:
        return self._state.btn_size

    # --------------------------------------------------------------------------
    # Qt method overrides
    # --------------------------------------------------------------------------
    def sizeHint(self) -> QSize:
        return QSize(self.x_size * self.btn_size, self.y_size * self.btn_size)

    @_filter_left_and_right
    def mousePressEvent(self, event: QMouseEvent):
        """Handle mouse press events."""
        coord = self._coord_from_event(event)

        # If the button pressed here is the only button down, we reset the
        # tracking variables.
        # However, if the click was outside the board, we should wait for all
        # buttons to be released before acknowledging clicks.
        if event.button() == event.buttons():
            self._both_mouse_buttons_pressed = False
            self._await_release_all_buttons = False
            self._was_double_left_click = False
            if not coord:
                self._await_release_all_buttons = True

        if self._was_double_left_click:
            self._await_release_all_buttons = True
            return
        if self._await_release_all_buttons:
            return

        self._mouse_coord = coord

        ## Bothclick
        if event.buttons() == (Qt.LeftButton | Qt.RightButton):
            logger.debug("Both mouse buttons down on cell %s", coord)
            self._both_mouse_buttons_pressed = True
            if coord:
                self.both_buttons_down(coord)
        ## Leftclick
        elif event.button() == Qt.LeftButton:
            logger.debug("Left mouse button down on cell %s", coord)
            assert coord is not None
            self.left_button_down(coord)
        ## Rightclick
        elif event.button() == Qt.RightButton:
            logger.debug("Right mouse button down on cell %s", coord)
            assert coord is not None
            self.right_button_down(coord)

        # The underlying coord may have changed.
        self._mouse_coord = self._coord_from_event(event)

    @_filter_left_and_right
    def mouseDoubleClickEvent(self, event: QMouseEvent):
        """Handle double clicks."""
        self._mouse_coord = coord = self._coord_from_event(event)

        if (
            event.button() == Qt.LeftButton
            and not self._both_mouse_buttons_pressed
            and coord
        ):
            self._was_double_left_click = True
            self.left_button_double_down(coord)
        else:
            return self.mousePressEvent(event)

        # The underlying coord may have changed.
        self._mouse_coord = self._coord_from_event(event)

    @_filter_left_and_right
    def mouseMoveEvent(self, event: QMouseEvent):
        """Handle mouse move events."""
        coord = self._coord_from_event(event)

        # Return if the mouse wasn't moved to a different cell.
        if self._await_release_all_buttons or coord == self._mouse_coord:
            return

        self._mouse_coord = coord

        ## Double leftclick
        if self._was_double_left_click:
            if event.buttons() == Qt.LeftButton:
                self.left_button_double_move(coord)
            return

        ## Bothclick
        if event.buttons() & Qt.LeftButton and event.buttons() & Qt.RightButton:
            self.both_buttons_move(coord)
        elif not self._both_mouse_buttons_pressed or self._state.drag_select:
            ## Leftclick
            if event.buttons() & Qt.LeftButton:
                self.left_button_move(coord)
            ## Rightclick
            if event.buttons() & Qt.RightButton:
                self.right_button_move(coord)

        # The underlying coord may have changed.
        self._mouse_coord = self._coord_from_event(event)

    @_filter_left_and_right
    def mouseReleaseEvent(self, event: QMouseEvent):
        """Handle mouse release events."""
        if self._await_release_all_buttons and not event.buttons():
            self._await_release_all_buttons = False
            return

        coord = self._coord_from_event(event)

        ## Bothclick (one of the buttons still down)
        if self._both_mouse_buttons_pressed:
            if event.buttons():
                logger.debug("Mouse button release on cell %s after both down", coord)
                self.first_of_both_buttons_release(coord)
            if not self._state.drag_select or event.buttons() == Qt.RightButton:
                # Only right button down - no risk.
                self.no_risk_signal.emit()
        ## Left release
        elif event.button() == Qt.LeftButton and not self._was_double_left_click:
            logger.debug("Left mouse button release on cell %s", coord)
            self.left_button_release(coord)

        # Reset variables if neither of the mouse buttons are down.
        if not event.buttons():
            logger.debug("No mouse buttons down, reset variables")
            self.all_buttons_release()

        # The underlying coord may have changed.
        self._mouse_coord = self._coord_from_event(event)

    # --------------------------------------------------------------------------
    # Mouse click handlers
    # --------------------------------------------------------------------------
    def left_button_down(self, coord: Coord) -> None:
        """
        Left mouse button was pressed (single click). Change display and call
        callback functions as appropriate.
        """
        if self._state.drag_select:
            self.at_risk_signal.emit()
            self._ctrlr.select_cell(coord)
        else:
            self._sink_unclicked_cells([coord])

    def left_button_double_down(self, coord: Coord) -> None:
        """
        Left button was double clicked. Call callback to remove any flags that
        were on the cell.
        """
        if self._board[coord].is_mine_type():
            self._ctrlr.remove_cell_flags(coord)
        else:
            self._was_double_left_click = False
            self.left_button_down(coord)

    def left_button_move(self, coord: Optional[Coord]) -> None:
        """
        Left mouse button was moved after a single click. Change display as
        appropriate.
        """
        self._raise_all_sunken_cells()
        self.no_risk_signal.emit()
        if coord is not None:
            self.left_button_down(coord)

    def left_button_double_move(self, coord: Optional[Coord]) -> None:
        """
        Left mouse button moved after a double click.
        """
        if self._state.drag_select and coord is not None:
            self._ctrlr.remove_cell_flags(coord)

    def left_button_release(self, coord: Coord) -> None:
        """
        Left mouse button was released. Change display and call callback
        functions as appropriate.
        """
        self._raise_all_sunken_cells()
        self.no_risk_signal.emit()
        if not self._state.drag_select and coord is not None:
            self._ctrlr.select_cell(coord)

    def right_button_down(self, coord: Coord) -> None:
        """
        Right mouse button was pressed. Change display and call callback
        functions as appropriate.
        """
        self._right_drag_action = self._impl.right_down_action(coord)

    def right_button_move(self, coord: Optional[Coord]) -> None:
        """
        Right mouse button was moved. Change display as appropriate.
        """
        if self._state.drag_select and coord is not None:
            self._impl.right_drag_action(coord, self._right_drag_action)

    def both_buttons_down(self, coord: Coord) -> None:
        """
        Both left and right mouse buttons were pressed. Change display and call
        callback functions as appropriate.
        """
        if not self._board[coord].is_mine_type():
            self._sink_unclicked_cells(self._board.get_nbrs(coord, include_origin=True))
        if self._state.drag_select:
            self.at_risk_signal.emit()
            self._ctrlr.chord_on_cell(coord)

    def both_buttons_move(self, coord: Optional[Coord]) -> None:
        """
        Both left and right mouse buttons were moved. Change display as
        appropriate.
        """
        self._raise_all_sunken_cells()
        self.no_risk_signal.emit()
        if coord is not None:
            self.both_buttons_down(coord)

    def first_of_both_buttons_release(self, coord: Coord) -> None:
        """
        One of the mouse buttons was released after both were pressed. Change
        display and call callback functions as appropriate.
        """
        self._raise_all_sunken_cells()
        if not self._state.drag_select and coord is not None:
            self._ctrlr.chord_on_cell(coord)

    def all_buttons_release(self) -> None:
        """
        The second of the mouse buttons was released after both were pressed.
        Change display and call callback functions as appropriate.
        """
        self.no_risk_signal.emit()
        self._mouse_coord = None
        self._both_mouse_buttons_pressed = False
        self._await_release_all_buttons = False
        self._was_double_left_click = False
        self._right_drag_action = FlagAction.FLAG

    # --------------------------------------------------------------------------
    # Other methods
    # --------------------------------------------------------------------------
    def _coord_from_event(self, event: QMouseEvent) -> Optional[Coord]:
        """
        Get cell coordinate from mouse button event.

        :param event:
            The mouse event.
        :return:
            The cell coordinate, or None if outside the board.
        """
        pos = self.mapToScene(event.pos())

        try:
            return self._board.get_coord_at(
                int(pos.x()) // self.btn_size, int(pos.y()) // self.btn_size
            )
        except ValueError:
            return None

    def _sink_unclicked_cells(self, coords: Iterable[Coord]) -> None:
        """
        Set an unclicked cell to appear sunken.
        """
        if self._state.game_status.finished():
            return
        sink_cells = {c for c in coords if self._board[c] is CellContents.Unclicked}
        if sink_cells:
            self._mouse_events.append(
                (self._elapsed, {c: SUNKEN_CELL for c in sink_cells})
            )
            self.at_risk_signal.emit()
            for c in sink_cells:
                self._set_cell_image(c, SUNKEN_CELL)
                self._sunken_cells.add(c)

    def _raise_all_sunken_cells(self) -> None:
        """
        Reset all sunken cells to appear raised.
        """
        raise_cells = {
            c for c in self._sunken_cells if self._board[c] is CellContents.Unclicked
        }
        if raise_cells:
            self._mouse_events.append(
                (self._elapsed, {c: RAISED_CELL for c in raise_cells})
            )
            for c in raise_cells:
                self._set_cell_image(c, RAISED_CELL)
                # TODO: HACK - fix hit mines in small cells in drag-select mode
                #  when using chording.
                if self._state.game_mode is GameMode.SPLIT_CELL and not c.is_split:
                    for small in c.get_small_cell_coords():
                        if small in self._board:
                            self._set_cell_image(small, self._board[small])
        self._sunken_cells.clear()

    def _set_cell_image(self, coord: Coord, state: CellContents) -> None:
        """Set the image of a cell."""
        self._impl.set_cell_image(coord, state)

    def _update_cell_images(self, img_type: CellImageType = CellImageType.ALL) -> None:
        self._impl.update_cell_images(img_type)

    def _update_size(self) -> None:
        self.setMaximumSize(self.sizeHint())
        self.setSceneRect(
            0, 0, self.x_size * self.btn_size, self.y_size * self.btn_size
        )
        self.size_changed.emit()

    def _redraw_cells(self) -> None:
        self._scene.clear()
        for coord in self._board.all_coords:
            self._set_cell_image(coord, self._board[coord])

    def switch_mode(self, mode: GameMode) -> None:
        self._impl = _IMPLS[mode](self._scene, self._ctrlr, self._state)

    def reset(self) -> None:
        """Reset all cell images and other state for a new game."""
        logger.info("Resetting minefield widget")
        self._redraw_cells()
        self._mouse_coord = None
        self._both_mouse_buttons_pressed = False
        self._await_release_all_buttons = True
        self._mouse_tracking = []
        self._mouse_events = []
        self._first_click_time = None

    def update_cells(self, cell_updates: Mapping[Coord, CellContents]) -> None:
        """
        Called to indicate some cells have changed state.

        :param cell_updates:
            A mapping of cell coordinates to their new state.
        """
        self._mouse_events.append((self._elapsed, cell_updates))
        for c, state in cell_updates.items():
            self._set_cell_image(c, state)
        # Always display cell updates as soon as possible.
        QApplication.processEvents()

    def reshape(self, x_size: int, y_size: int) -> None:
        logger.info("Resizing minefield to %sx%s", x_size, y_size)
        self._update_size()
        self._redraw_cells()

    def update_style(self, img_type: CellImageType, style: str) -> None:
        """Update the cell images."""
        logger.info("Updating %s style to '%s'", img_type.name, style)
        self._update_cell_images(img_type)
        self._redraw_cells()

    def update_btn_size(self, size: int) -> None:
        """Update the size of the cells."""
        assert size == self._state.btn_size
        self._update_cell_images()
        self._redraw_cells()
        self._update_size()

    def get_mouse_events(self) -> List[CellUpdate_T]:
        return self._mouse_events



# November 2018, Lewis Gaul

"""
The core game engine.

"""

__all__ = ("UberController",)

import json
import logging
import sys
from typing import Mapping, Type

from ..shared.types import Coord, Difficulty, GameMode, PathLike, UIMode
from ..shared.utils import GameOptsStruct
from . import api, board, controller, game, minefield, regular, split_cell
from .board import BoardBase
from .controller import ControllerBase


if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol


logger = logging.getLogger(__name__)


class GameModeImplementation(Protocol):
    """Protocol for game mode implementations to satisfy."""

    Minefield: Type[minefield.MinefieldBase]
    Board: Type[board.BoardBase]
    Game: Type[game.GameBase]
    GameController: Type[controller.ControllerBase]
    CreateController: Type[controller.ControllerBase]
    mode: GameMode


GAME_MODE_IMPL: Mapping[GameMode, GameModeImplementation] = {
    GameMode.REGULAR: regular,
    GameMode.SPLIT_CELL: split_cell,
}


class UberController(api.AbstractController):
    """Base controller implementing all user interaction methods."""

    def __init__(self, opts: GameOptsStruct):
        super().__init__(opts)
        self._ui_mode: UIMode = UIMode.GAME
        self._active_ctrlr: ControllerBase = self._get_ctrlr_cls(
            self.mode, self._ui_mode
        )(self._opts, notif=self._notif)

    @property
    def mode(self) -> GameMode:
        return self._opts.mode

    @staticmethod
    def _get_ctrlr_cls(game_mode: GameMode, ui_mode: UIMode) -> Type[ControllerBase]:
        """Get the controller class for given modes."""
        if ui_mode is UIMode.GAME:
            return GAME_MODE_IMPL[game_mode].GameController
        elif ui_mode is UIMode.CREATE:
            return GAME_MODE_IMPL[game_mode].CreateController
        else:
            raise ValueError(f"Unsupported UI mode {ui_mode}")

    def switch_game_mode(self, mode: GameMode) -> None:
        """Switch the game mode."""
        super().switch_game_mode(mode)
        if mode is self.mode:
            logger.debug(
                "Ignore switch game mode request because mode is already %s", mode
            )
            return
        if mode is GameMode.SPLIT_CELL and (
            self._opts.x_size % 2 != 0 or self._opts.y_size % 2 != 0
        ):
            self.resize_board(
                self._opts.x_size // 2 * 2, self._opts.y_size // 2 * 2, self._opts.mines
            )
        self._notif.game_mode_about_to_change(mode)
        self._opts.mode = mode
        old_difficulty = self._active_ctrlr.difficulty
        self._active_ctrlr = self._get_ctrlr_cls(mode, self._ui_mode)(
            self._opts, notif=self._notif
        )
        if old_difficulty is not Difficulty.CUSTOM:
            self.set_difficulty(old_difficulty)
        else:
            new_difficulty = self._active_ctrlr.game_cls.difficulty_from_values(
                self._opts.x_size, self._opts.y_size, self._opts.mines
            )
            if new_difficulty is not Difficulty.CUSTOM:
                self._notif.set_difficulty(new_difficulty)
        self._notif.game_mode_changed(mode)

    def switch_ui_mode(self, ui_mode: UIMode) -> None:
        """Switch the UI mode."""
        super().switch_ui_mode(ui_mode)
        if ui_mode is self._ui_mode:
            logger.debug(
                "Ignore switch UI mode request because mode is already %s", ui_mode
            )
            return
        self._ui_mode = ui_mode
        self._active_ctrlr = self._get_ctrlr_cls(self.mode, ui_mode)(
            self._opts, notif=self._notif
        )
        self._notif.reset()

    def reset_settings(self) -> None:
        super().reset_settings()
        self._opts = GameOptsStruct()
        self.switch_ui_mode(UIMode.GAME)
        self.resize_board(self._opts.x_size, self._opts.y_size, self._opts.mines)
        self._notif.reset()

    def load_minefield(self, file: PathLike) -> None:
        """Load a minefield from file."""
        super().load_minefield(file)

        if self._ui_mode is UIMode.CREATE:
            self.switch_ui_mode(UIMode.GAME)
            self._notif.ui_mode_changed(UIMode.GAME)

        with open(file, mode="r", encoding="utf-8") as f:
            data = json.load(f)

        try:
            game_mode = GameMode(data.pop("type").upper())
        except (KeyError, ValueError):
            game_mode = GameMode.REGULAR
        if game_mode is not self.mode:
            self.switch_game_mode(game_mode)

        self._active_ctrlr.load_minefield(file)

    # ----------------------------------
    # Delegated abstractmethods
    # ----------------------------------
    @property
    def board(self) -> BoardBase:
        return self._active_ctrlr.board

    def get_game_info(self) -> api.GameInfo:
        return self._active_ctrlr.get_game_info()

    def new_game(self) -> None:
        self._active_ctrlr.new_game()

    def restart_game(self) -> None:
        self._active_ctrlr.restart_game()

    def select_cell(self, coord: Coord) -> None:
        self._active_ctrlr.select_cell(coord)

    def flag_cell(self, coord: Coord, *, flag_only: bool = False) -> None:
        self._active_ctrlr.flag_cell(coord, flag_only=flag_only)

    def split_cell(self, coord: Coord) -> None:
        self._active_ctrlr.split_cell(coord)

    def chord_on_cell(self, coord: Coord) -> None:
        self._active_ctrlr.chord_on_cell(coord)

    def remove_cell_flags(self, coord: Coord) -> None:
        self._active_ctrlr.remove_cell_flags(coord)

    def resize_board(self, x_size: int, y_size: int, mines: int) -> None:
        self._active_ctrlr.resize_board(x_size, y_size, mines)

    def set_difficulty(self, difficulty: Difficulty) -> None:
        self._active_ctrlr.set_difficulty(difficulty)

    def set_first_success(self, value: bool) -> None:
        self._active_ctrlr.set_first_success(value)

    def set_per_cell(self, value: int) -> None:
        self._active_ctrlr.set_per_cell(value)

    def save_current_minefield(self, file: PathLike) -> None:
        self._active_ctrlr.save_current_minefield(file)



# September 2019, Lewis Gaul

"""
API between backend and frontends.

"""

__all__ = (
    "AbstractController",
    "AbstractListener",
    "GameInfo",
)

import abc
import logging
from typing import Callable, Dict, Iterable, List, Optional

import attr

from ..shared.types import (
    CellContents,
    Coord,
    Difficulty,
    GameMode,
    GameState,
    PathLike,
    UIMode,
)
from ..shared.utils import GameOptsStruct
from .board import BoardBase


@attr.attrs(auto_attribs=True, kw_only=True)
class GameInfo:
    """General information about a game."""

    @attr.attrs(auto_attribs=True, kw_only=True)
    class StartedInfo:
        start_time: float
        elapsed: float
        bbbv: int
        rem_bbbv: int
        bbbvps: float
        prop_complete: float
        prop_flagging: float

    game_state: GameState
    x_size: int
    y_size: int
    mines: int
    difficulty: Difficulty
    per_cell: int
    first_success: bool
    mode: GameMode

    minefield_known: bool
    started_info: Optional[StartedInfo] = None


class AbstractListener(metaclass=abc.ABCMeta):
    """
    An abstract class outlining methods that should be implemented to receive
    updates on changes to state. Instances of a concrete implementation can
    then be registered to listen for callbacks.
    """

    @abc.abstractmethod
    def reset(self) -> None:
        """
        Called to indicate the state should be reset.
        """
        return NotImplemented

    @abc.abstractmethod
    def resize_minefield(self, x_size: int, y_size: int) -> None:
        """
        Called to indicate the board is being changed.

        :param x_size:
            The number of columns.
        :param y_size:
            The number of rows.
        """
        return NotImplemented

    @abc.abstractmethod
    def set_mines(self, mines: int) -> None:
        """Called to indicate the number of base mines has changed."""
        return NotImplemented

    @abc.abstractmethod
    def set_difficulty(self, diff: Difficulty) -> None:
        """Called to indicate the difficulty has changed."""
        return NotImplemented

    @abc.abstractmethod
    def update_cells(self, cell_updates: Dict[Coord, CellContents]) -> None:
        """
        Called when one or more cells were updated.

        :param cell_updates:
            Mapping of coordinates that were changed to the new cell state.
        """
        return NotImplemented

    @abc.abstractmethod
    def update_game_state(self, game_state: GameState) -> None:
        """
        Called when the game state changes.

        :param game_state:
            The new game state.
        """
        return NotImplemented

    @abc.abstractmethod
    def update_mines_remaining(self, mines_remaining: int) -> None:
        """
        Called when the number of mines remaining changes.

        :param mines_remaining:
            The new number of mines remaining.
        """
        return NotImplemented

    @abc.abstractmethod
    def ui_mode_changed(self, mode: UIMode) -> None:
        """
        Called to indicate the UI mode has changed.

        :param mode:
            The mode to change to.
        """
        return NotImplemented

    @abc.abstractmethod
    def game_mode_about_to_change(self, mode: GameMode) -> None:
        """
        Called to indicate the game mode is about to change.

        :param mode:
            The mode to change to.
        """
        return NotImplemented

    @abc.abstractmethod
    def game_mode_changed(self, mode: GameMode) -> None:
        """
        Called to indicate the game mode has changed.

        :param mode:
            The mode to change to.
        """
        return NotImplemented

    @abc.abstractmethod
    def handle_exception(self, method: str, exc: Exception) -> None:
        """
        Called if an exception occurs when calling any of the other methods on
        the class, to allow the implementer of the class to handle (e.g. log)
        any errors.

        :param method:
            The method that the exception occurred in.
        :param exc:
            The caught exception.
        """
        return NotImplemented


class _Notifier(AbstractListener):
    """Pass on calls to registered listeners."""

    _count: int = 0

    def __init__(self, listeners: Iterable[AbstractListener] = None):
        """
        Create the implementation for all

        :param listeners:
        """
        self._listeners: List[AbstractListener] = list(listeners) if listeners else []
        self._id: int = self._count
        self._logger = logging.getLogger(
            f"{__name__}.{self.__class__.__name__}[{self._id}]"
        )

        self.__class__._count += 1

        # Do the method wrapping here because we need the registered listeners.
        for method in AbstractListener.__abstractmethods__:
            setattr(self, method, self._call_registered(method))

    def register_listener(self, listener: AbstractListener) -> None:
        """
        Register a listener to receive updates from the controller.

        :param listener:
            An AbstractListener instance to register.
        """
        self._listeners.append(listener)

    def unregister_listener(self, listener: AbstractListener) -> None:
        """
        Unregister a listener to receive updates from the controller.

        Does nothing if not registered.

        :param listener:
            An AbstractListener instance to unregister.
        """
        try:
            self._listeners.remove(listener)
        except ValueError:
            self._logger.debug("Listener not registered - nothing to do")

    def _call_registered(self, func: str) -> Callable:
        """
        Decorator to call all registered listeners.

        :param func:
            The name of the method to decorate.
        :return:
            The decorated version of the method.
        """
        if not hasattr(self, func + "_orig"):
            setattr(self, func + "_orig", getattr(self, func))

        def wrapped(*args, **kwargs):
            getattr(self, func + "_orig")(*args, **kwargs)
            for listener in self._listeners:
                try:
                    getattr(listener, func)(*args, **kwargs)
                except Exception as e:
                    self._logger.warning(
                        f"Error occurred calling {func}() on {listener}"
                    )
                    listener.handle_exception(func, e)

        return wrapped

    def reset(self) -> None:
        """
        Called to indicate the state should be reset.
        """
        self._logger.debug("Calling reset()")

    def resize_minefield(self, x_size: int, y_size: int) -> None:
        """
        Called to indicate the board shape has changed.

        :param x_size:
            The number of rows.
        :param y_size:
            The number of columns.
        """
        self._logger.debug("Calling resize_minefield() with %d, %d", x_size, y_size)

    def set_mines(self, mines: int) -> None:
        self._logger.debug("Calling set_mines() with %d", mines)

    def set_difficulty(self, diff: Difficulty) -> None:
        self._logger.debug("Calling set_difficulty() with %s", diff)

    def update_cells(self, cell_updates: Dict[Coord, CellContents]) -> None:
        """
        Called when one or more cells were updated.

        :param cell_updates:
            Mapping of coordinates that were changed to the new cell state.
        """
        self._logger.debug(
            f"Calling update_cells() with {len(cell_updates)} updated cells"
        )

    def update_game_state(self, game_state: GameState) -> None:
        """
        Called when the game state changes.

        :param game_state:
            The new game state.
        """
        self._logger.debug(f"Calling update_game_state() with {game_state}")

    def update_mines_remaining(self, mines_remaining: int) -> None:
        """
        Called when the number of mines remaining changes.

        :param mines_remaining:
            The new number of mines remaining.
        """
        self._logger.debug(f"Calling update_mines_remaining() with {mines_remaining}")

    def ui_mode_changed(self, mode: UIMode) -> None:
        """
        Called to indicate the UI mode has changed.

        :param mode:
            The mode to change to.
        """
        self._logger.debug("Calling ui_mode_changed() with %r", mode.name)

    def game_mode_about_to_change(self, mode: GameMode) -> None:
        """
        Called to indicate the game mode is about to change.

        :param mode:
            The mode to change to.
        """
        self._logger.debug("Calling game_mode_about_to_change() with %r", mode.name)

    def game_mode_changed(self, mode: GameMode) -> None:
        """
        Called to indicate the game mode has changed.

        :param mode:
            The mode to change to.
        """
        self._logger.debug("Calling game_mode_changed() with %r", mode.name)

    def handle_exception(self, method: str, exc: Exception) -> None:
        """
        Not used in this class - provided only to satisfy the ABC.
        """
        return NotImplemented


class AbstractController(metaclass=abc.ABCMeta):
    """
    Abstract controller base class.

    Listeners can be registered for receiving updates.
    """

    def __init__(self, opts: GameOptsStruct):
        self._opts = GameOptsStruct.from_structs(opts)
        # The registered functions to be called with updates.
        self._notif = _Notifier()
        self._logger = logging.getLogger(
            f"{self.__class__.__module__}.{self.__class__.__name__}"
        )

    def register_listener(self, listener: AbstractListener) -> None:
        """
        Register a listener to receive updates from the controller.

        :param listener:
            An AbstractListener instance to register.
        """
        self._logger.info(
            "Registering listener: %s.%s",
            type(listener).__module__,
            type(listener).__name__,
        )
        self._notif.register_listener(listener)
        # TODO Implement in subclass to update with current state.

    def unregister_listener(self, listener: AbstractListener) -> None:
        """
        Unregister a listener to receive updates from the controller.

        Does nothing if not registered.

        :param listener:
            An AbstractListener instance to unregister.
        """
        self._logger.info(
            "Unregistering listener: %s.%s",
            type(listener).__module__,
            type(listener).__name__,
        )
        self._notif.unregister_listener(listener)

    # --------------------------------------------------------------------------
    # Getters
    # --------------------------------------------------------------------------
    @property
    @abc.abstractmethod
    def board(self) -> BoardBase:
        return NotImplemented

    @abc.abstractmethod
    def get_game_info(self) -> GameInfo:
        """Get information about the current game."""
        return NotImplemented

    def get_game_options(self) -> GameOptsStruct:
        return self._opts

    # --------------------------------------------------------------------------
    # Methods triggered by user interaction
    # --------------------------------------------------------------------------
    @abc.abstractmethod
    def new_game(self) -> None:
        """
        Create a new game, refresh the board state.
        """
        self._logger.info("New game requested, refreshing the board")

    @abc.abstractmethod
    def restart_game(self) -> None:
        """
        Restart the current game, refresh the board state.
        """
        self._logger.info("Restart game requested, refreshing the board")

    @abc.abstractmethod
    def select_cell(self, coord: Coord) -> None:
        """
        Select a cell for a regular click.
        """
        self._logger.debug("Cell %s selected", coord)

    @abc.abstractmethod
    def flag_cell(self, coord: Coord, *, flag_only: bool = False) -> None:
        """Select a cell for flagging."""
        self._logger.debug("Cell %s selected for flagging", coord)

    def split_cell(self, coord: Coord) -> None:
        """Split a cell - only required for split-cell mode."""
        self._logger.debug("Cell %s selected to be split", coord)

    @abc.abstractmethod
    def chord_on_cell(self, coord: Coord) -> None:
        """
        Select a cell for chording.
        """
        self._logger.debug("Cell %s selected for chording", coord)

    @abc.abstractmethod
    def remove_cell_flags(self, coord: Coord) -> None:
        """
        Remove flags in a cell, if any.
        """
        self._logger.debug("Flags in cell %s being removed", coord)

    @abc.abstractmethod
    def resize_board(self, x_size: int, y_size: int, mines: int) -> None:
        """
        Resize the board and/or change the number of mines.
        """
        self._logger.info(
            "Resizing the board to %sx%s with %s mines", x_size, y_size, mines
        )

    @abc.abstractmethod
    def set_difficulty(self, difficulty: Difficulty) -> None:
        """
        Set the size of the board and the number of mines for the given difficulty.
        """
        self._logger.info("Changing the board to difficulty %s", difficulty.name)

    @abc.abstractmethod
    def set_first_success(self, value: bool) -> None:
        """
        Set whether the first click should be a guaranteed success.
        """
        self._logger.debug("Setting first success to %s", value)

    @abc.abstractmethod
    def set_per_cell(self, value: int) -> None:
        """
        Set the maximum number of mines per cell.
        """
        self._logger.debug("Setting per cell to %s", value)

    @abc.abstractmethod
    def save_current_minefield(self, file: PathLike) -> None:
        """
        Save the current minefield to file.

        :param file:
            The location of the file to save to. Should have the extension
            ".mgb".
        """
        self._logger.debug("Saving current minefield to file: %s", file)

    @abc.abstractmethod
    def load_minefield(self, file: PathLike) -> None:
        """
        Load a minefield from file.

        :param file:
            The location of the file to load from. Should have the extension
            ".mgb".
        """
        self._logger.debug("Loading minefield from file: %s", file)

    @abc.abstractmethod
    def switch_game_mode(self, mode: GameMode) -> None:
        """
        Switch the game mode, e.g. into 'split cells' mode.
        """
        self._logger.info("Requested switch to game mode %s", mode)

    @abc.abstractmethod
    def switch_ui_mode(self, mode: UIMode) -> None:
        """
        Switch the UI mode, e.g. into 'create' mode.
        """
        self._logger.info("Requested switch to UI mode %s", mode)

    @abc.abstractmethod
    def reset_settings(self) -> None:
        """
        Reset all settings to the defaults.
        """
        self._logger.info("Requested reset of settings")



# March 2018, Lewis Gaul

"""
Minesweeper board API.

"""

__all__ = ("BoardBase",)

import abc
from typing import Iterable, List

from ..shared.types import CellContents, Coord, GameMode


class BoardBase(metaclass=abc.ABCMeta):
    """Representation of a minesweeper board, generic over the game mode."""

    mode: GameMode

    @abc.abstractmethod
    def __eq__(self, other) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def __getitem__(self, coord: Coord) -> CellContents:
        raise NotImplementedError

    @abc.abstractmethod
    def __setitem__(self, coord: Coord, contents: CellContents) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def __contains__(self, coord: Coord) -> bool:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def all_coords(self) -> List[Coord]:
        """A list of all coords currently in the board."""
        raise NotImplementedError

    @property
    def all_underlying_coords(self) -> List[Coord]:
        """A list of all underlying coords that may contain a mine."""
        return self.all_coords

    @abc.abstractmethod
    def get_nbrs(
        self, coord: Coord, *, include_origin: bool = False
    ) -> Iterable[Coord]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_coord_at(self, x: int, y: int) -> Coord:
        raise NotImplementedError
